(*<*)
theory RHL = VS:
(*>*)

section{*Termination-insensitive RHL*}

text{*We first define the type of state relations and the
termination-insensitive interpretation of relational judgements in the
case where the two programs are identical.*}

types REL = "(State \<times> State) set"

constdefs RELsecure::"REL \<Rightarrow> IMP \<Rightarrow> REL \<Rightarrow> bool"
"RELsecure R C S \<equiv> \<forall> s t ss tt . (s, t):R \<longrightarrow> (s,C \<Down> ss) \<longrightarrow> 
                          (t,C \<Down> tt) \<longrightarrow> (ss, tt):S"

text {*Here is the definition of the assertion transformer
$\mathit{Sec}$ \ldots*}

constdefs RELSec :: "REL \<Rightarrow> REL \<Rightarrow> TT \<Rightarrow> VDMAssn"
"RELSec R S phi s t \<equiv> (\<forall> r . (s, r):R \<longrightarrow> phi(t, r)) \<and> 
             (\<forall> r . phi(r ,s) \<longrightarrow> (r, t):S)"

text{*\ldots and the proofs of its characteristic property,
Proposition 5.*}

lemma RELProp5A : "vdm_valid C (RELSec R S Phi) \<Longrightarrow> RELsecure R C S"
by (simp add: vdm_valid_def RELsecure_def RELSec_def)


lemma RELProp5B:
  "RELsecure R C S \<Longrightarrow> vdm_valid C (RELSec R S (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))"
(*<*)
apply (simp add: vdm_valid_def RELSec_def) 
apply clarsimp
apply (unfold RELsecure_def)
apply rule 
apply (rule, rule) apply (rule_tac x=s in exI, rule, assumption+)
apply (rule, rule, erule exE, erule conjE) apply fast
done
(*>*)

lemma RELProp5BB:"RELsecure R C S \<Longrightarrow> \<exists> Phi . vdm_valid C (RELSec R S Phi)"
(*<*)
by (drule RELProp5B, fast)
(*>*)

lemma RELProp5:"(RELsecure R C S)
             = (vdm_valid C (RELSec R S (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R)))"
(*<*)
apply rule
apply (erule RELProp5B)
apply (erule RELProp5A)
done
(*>*)

text{*Next, the definitions of RSecL and RSecR.*}

constdefs RELSecL :: "REL \<Rightarrow> TT \<Rightarrow> VDMAssn"
"RELSecL R phi s t \<equiv> \<forall> r . (s, r):R \<longrightarrow> phi(t, r)"

constdefs RELSecR :: "REL \<Rightarrow> TT \<Rightarrow> VDMAssn"
"RELSecR S phi s t \<equiv> \<forall> r . phi(r ,s) \<longrightarrow> (r, t):S"

lemma RELSec_RELSecL_RELSecR:
"RELSec R S phi = INTER (RELSecL R phi) (RELSecR S phi)"
(*<*)
apply (rule, rule)
apply (simp add: INTER_def RELSec_def RELSecL_def RELSecR_def) 
done
(*>*)

text{*Here's the definition of the termination-insensitive
interpretation in case where the programs are different.*}

constdefs BINsecure::"REL \<Rightarrow> IMP \<Rightarrow> IMP \<Rightarrow> REL \<Rightarrow> bool"
"BINsecure R C D S \<equiv> \<forall> s t ss tt . (s, t):R \<longrightarrow> (s,C \<Down> ss) \<longrightarrow> 
                          (t,D \<Down> tt) \<longrightarrow> (ss, tt):S"

lemma BINsecure_RELsecure:
"BINsecure R C C S = RELsecure R C S"
(*<*)
by (simp add: BINsecure_def RELsecure_def)
(*>*)

lemma BINProp5A_separated : 
"\<lbrakk>vdm_valid C (RELSecL R Phi); vdm_valid D (RELSecR S Phi)\<rbrakk> 
\<Longrightarrow> BINsecure R C D S"
(*<*)
by (simp add: vdm_valid_def BINsecure_def RELSecL_def RELSecR_def)
(*>*)

text{*From this we obtain the simplified Prop5A as used above:*}

lemma BINProp5A : 
"\<lbrakk>vdm_valid C (RELSec R S Phi); vdm_valid D (RELSec R S Phi)\<rbrakk> 
\<Longrightarrow> BINsecure R C D S"
(*>*)
apply (rule BINProp5A_separated [of C R Phi D S])
apply (simp add: vdm_valid_def RELSec_def RELSecL_def)
apply (simp add: vdm_valid_def RELSec_def RELSecR_def)
done
(*>*)

text{*Of course, this result can also be proven directly:*}
lemma
"\<lbrakk>vdm_valid C (RELSec R S Phi); vdm_valid D (RELSec R S Phi)\<rbrakk> 
\<Longrightarrow> BINsecure R C D S"
(*<*)
by (simp add: vdm_valid_def BINsecure_def RELSec_def)
(*>*)

lemma BINProp5B_left:
  "BINsecure R C D S \<Longrightarrow> vdm_valid C (RELSecL R (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))"
(*<*)
apply (simp add: vdm_valid_def RELSecL_def) 
apply clarsimp
apply (unfold BINsecure_def)
apply (rule_tac x=s in exI, rule, assumption+)
done
(*>*)

lemma BINProp5B_right:
  "BINsecure R C D S \<Longrightarrow> vdm_valid D (RELSecR S (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))"
(*<*)
apply (simp add: vdm_valid_def RELSecR_def) 
apply clarsimp
apply (unfold BINsecure_def)
apply fast
done
(*>*)

lemma BINProp5B:
  "BINsecure R C D S \<Longrightarrow> 
     vdm_valid C (RELSecL R (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R)) \<and>
     vdm_valid D (RELSecR S (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))"
(*<*)
apply rule
apply (erule BINProp5B_left)
apply (erule BINProp5B_right)
done
(*>*)

lemma BINProp5BB:"BINsecure R C D S \<Longrightarrow> 
\<exists> Phi . vdm_valid C (RELSecL R Phi) \<and> vdm_valid D (RELSecR S Phi)"
(*<*)
by (drule BINProp5B, fast)
(*>*)

lemma BINProp5:
  "(BINsecure R C D S) = 
   ((vdm_valid C (RELSecL R (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))) \<and>
    (vdm_valid D (RELSecR S (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> (s, t):R))))"
(*<*)
apply rule
apply (erule BINProp5B)
apply (erule conjE) apply (erule BINProp5A_separated) apply assumption
done
(*>*)

text{*For proving the claim about symmetry in the paper we define the
inverse of a relation.*}

constdefs Inverse::"REL \<Rightarrow> REL"
"Inverse R \<equiv> {(x,y) . (y,x):R}"

lemma BINsecure_symmetric: 
"BINsecure R C D S = BINsecure (Inverse R) D C (Inverse S)"
(*<*)
by (simp add: BINsecure_def Inverse_def, fast)
(*>*)

lemma RHL_symm_semantic1:
"(\<exists> phi . vdm_valid C (RELSecL R phi) \<and> vdm_valid D (RELSecR S phi)) =
(\<exists> phi . vdm_valid D (RELSecL (Inverse R) phi) \<and> vdm_valid C (RELSecR (Inverse S) phi))"
(*<*)
apply rule
  apply clarsimp 
  apply (drule BINProp5A_separated, assumption) 
  apply (subgoal_tac "BINsecure (Inverse R) D C (Inverse S)")
    apply (erule thin_rl, erule thin_rl)
    apply (erule BINProp5BB)
  defer 
apply clarsimp
  apply (rule BINProp5BB) 
  apply (subgoal_tac "BINsecure (Inverse R) D C (Inverse S)")
    defer 
    apply (erule BINProp5A_separated) apply assumption
  apply (insert BINsecure_symmetric) apply fast apply fast
done
(*>*)

(*<*) (*A semantic proof for the composition rule*)
lemma BINsecure_relational_composition:
  "\<lbrakk>BINsecure R C D T; BINsecure U D E V;
        \<forall> s s' . (s,s'):R \<longrightarrow> (\<exists> t. s,C \<Down> t) \<longrightarrow> (\<exists> t'. s',D\<Down>t') \<rbrakk>
       \<Longrightarrow> BINsecure {(s,s'). \<exists> r . (s,r):R \<and> (r,s'):U} C E {(s,s'). \<exists> r . (s,r):T \<and> (r,s'):V}"
apply (simp add: BINsecure_def)
apply clarsimp
apply (erule_tac x=s in allE, erule_tac x=r in allE, erule impE, assumption)
apply (erule_tac x=ss in allE, erule impE, assumption)
apply (erule_tac x=s in allE, erule_tac x=r in allE, erule impE, assumption)
apply (erule impE) apply (rule_tac x=ss in exI, assumption) apply clarsimp
apply (erule_tac x=t' in allE, erule impE, assumption)
apply (erule_tac x=r in allE, erule_tac x=t in allE, erule impE, assumption)
apply (erule_tac x=t' in allE, erule impE, assumption)
apply (erule_tac x=tt in allE, erule impE, assumption)
apply (rule_tac x=t' in exI, simp)
done
(*>*)

constdefs RHL::"REL \<Rightarrow> IMP \<Rightarrow> IMP \<Rightarrow> REL \<Rightarrow> TT \<Rightarrow> bool"
"RHL R C D S Phi \<equiv> {} \<rhd> C : (RELSecL R Phi) \<and> {} \<rhd> D : (RELSecR S Phi)"

text{*By definition, and Prop1, we have the following.*}

lemma RHL_valid_implies_BIN_secure:
"RHL R C D S Phi \<Longrightarrow> BINsecure R C D S"
(*<*)
apply (unfold RHL_def) apply clarsimp
apply (drule VDM_Sound) apply (simp add: v_valid_def) 
  apply (erule impE, simp add: vdmctxt_valid_def)
apply (drule VDM_Sound) apply (simp add: v_valid_def) 
  apply (erule impE, simp add: vdmctxt_valid_def) 
apply (erule BINProp5A_separated, assumption)
done
(*>*)

text{*Thus, it is now our aim to derive Benton's rules. We separate
the proof of each rule into its RELSecL and RELSecR components.*}

lemma REL_UnaryL:
  "\<lbrakk> {} \<rhd> C : A; phi = (\<lambda> (s, t) . \<exists> u . A u s \<and> (u,t):R);
     R = {(s, s'). \<forall>t. A s t \<longrightarrow> (\<forall>t'. A' s' t' \<longrightarrow> (t, t') \<in> S)}\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecL R phi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecL_def) apply clarsimp
  apply (rule_tac x=s in exI, simp) 
done
(*>*)

lemma REL_UnaryR:
  "\<lbrakk>{} \<rhd> C' : A'; phi = (\<lambda>(s, t). \<exists>u. A u s \<and> (\<forall>ta. A u ta \<longrightarrow> (\<forall>t'. A' t t' \<longrightarrow> (ta, t') \<in> S)));
     R = {(s, s'). \<forall>t. A s t \<longrightarrow> (\<forall>t'. A' s' t' \<longrightarrow> (t, t') \<in> S)}\<rbrakk>
    \<Longrightarrow> {} \<rhd> C' : RELSecR S phi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecR_def) apply clarsimp
done
(*>*)

lemma REL_Unary:
  "\<lbrakk> {} \<rhd> C:A; {} \<rhd> C' : A'; phi = (\<lambda> (s, t) . \<exists> u . A u s \<and> (u,t):R);
     R = {(s,s') . \<forall> t t' . A s t \<longrightarrow> A' s' t' \<longrightarrow> (t,t'):S}\<rbrakk>
   \<Longrightarrow> RHL R C C' S phi" 
apply (simp add: RHL_def)
apply rule
  apply (erule REL_UnaryL) apply simp apply simp
  apply (erule REL_UnaryR) apply simp apply simp
done

text{*For RHL-symm we don't have the perferct separation - the
jdugement for $D$ does not need either of the two hypothesis, but the
justification of the judgement of $C$ needs both. This is due to the
unsymmetry in the definition of the Sec operators.*}

lemma RELSymmL:
"{} \<rhd> D : (RELSecL (Inverse R) (\<lambda>(r, t). \<exists>s. s , D \<Down> r  \<and> (s, t) \<in> Inverse R))"
(*<*)
(*with VDMOpsem,we could do this:
apply (rule VDMConseq) 
apply (rule VDMTrue) apply (simp add: RELSecL_def) apply clarsimp
apply (rule_tac x=s in exI, simp)*)
apply (rule VDM_Complete)
apply (simp add: vdm_valid_def) apply (simp add: RELSecL_def) apply clarsimp
apply (rule_tac x=s in exI, simp)
done
(*>*)

lemma RELSymmR:
  "\<lbrakk>{} \<rhd> D : (RELSecR S phi); {} \<rhd> C : (RELSecL R phi)\<rbrakk>
   \<Longrightarrow> {} \<rhd> C : (RELSecR (Inverse S) (\<lambda>(r, t). \<exists>s. s , D \<Down> r  \<and> (s, t) \<in> Inverse R))"
(*<*)
apply (rule VDM_Complete)
apply (simp add: vdm_valid_def) apply (simp add: RELSecR_def) apply clarsimp
apply (drule VDM_Sound)
apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def) apply (simp add: RELSecR_def Inverse_def)
apply (erule_tac x=sa in allE, erule_tac x=r in allE, clarsimp)
apply (erule_tac x=t in allE, erule mp)
apply (drule VDM_Sound)
apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def) apply (simp add: RELSecL_def)
done
(*>*)

lemma RHL_symm:
  "RHL R C D S phi \<Longrightarrow> RHL (Inverse R) D C (Inverse S) (\<lambda>(r, t). \<exists>s. s , D \<Down> r  \<and> (s, t) \<in> Inverse R)"
(*<*)
apply (simp add: RHL_def, erule conjE, rule)
apply (rule RELSymmL)
apply (erule RELSymmR) apply assumption
done
(*>*)

text{*The following alternative proof of this rule uses the semantic
validity of the rule and the completeness of the VDM Logic*}

lemma RHL_symm_alternative_proof:
  "RHL R C D S phi \<Longrightarrow> RHL (Inverse R) D C (Inverse S) (\<lambda>(r, t). \<exists>s. s , D \<Down> r  \<and> (s, t) \<in> Inverse R)"
(*<*)
apply (drule RHL_valid_implies_BIN_secure) 
apply (subgoal_tac "BINsecure (Inverse R) D C (Inverse S)")
  apply (erule thin_rl)
  apply (drule BINProp5B) apply (erule conjE)
  apply (simp add: RHL_def) 
  apply rule apply (erule VDM_Complete) apply (erule VDM_Complete)
apply (insert BINsecure_symmetric)apply fast
done
(*>*)

lemma BentonCBInvTL_L:
 "\<lbrakk>{} \<rhd> (Iff b C C') : RELSecL RR phi; R = {(s,s') . (s,s') : RR \<and> evalB b s}\<rbrakk>
 \<Longrightarrow> {} \<rhd> C : RELSecL R phi"
(*<*)
apply (drule VDM_Sound)
apply (rule VDM_Complete)
apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def) apply clarsimp
apply (simp add: RELSecL_def) apply clarsimp
apply (erule_tac x=s in allE, erule_tac x=t in allE)
apply (erule impE) apply (simp add: Sem_def, clarsimp)
  apply (rule, erule SemTrue)  apply assumption
apply simp
done
(*>*)

lemma BentonCBInvTL_R:
 "{} \<rhd>  D : RELSecR S phi 
 \<Longrightarrow> {} \<rhd> D : RELSecR S phi"
(*<*) by simp (*>*)

lemma BentonCBInvTL:
 "\<lbrakk>RHL RR (Iff b C C') D S phi; R = {(s,s') . (s,s') : RR \<and> evalB b s}\<rbrakk>
 \<Longrightarrow> RHL R C D S phi"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply (erule BentonCBInvTL_L) apply simp
done
(*>*)

lemma BentonCBInvTL_alternative_proof:
 "\<lbrakk>RHL RR (Iff b C C') D S phi; R = {(s,s') . (s,s') : RR \<and> evalB b s}\<rbrakk>
 \<Longrightarrow> RHL R C D S phi"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply (drule VDM_Sound)
apply (rule VDM_Complete)
apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def) apply clarsimp
apply (simp add: RELSecL_def) apply clarsimp
apply (erule_tac x=s in allE, erule_tac x=t in allE)
apply (erule impE) apply (simp add: Sem_def, clarsimp)
  apply (rule, erule SemTrue)  apply assumption
apply simp
done
(*>*)

constdefs SkipPhi::"REL \<Rightarrow> TT"
"SkipPhi R \<equiv> (\<lambda> (x,y) . (x,y):R)"

lemma REL_SkipL:"{} \<rhd> Skip : (RELSecL R (SkipPhi R))"
(*<*)
apply (rule VDMConseq) apply (rule VDMSkip)
  apply (simp add: RELSecL_def SkipPhi_def) 
done
(*>*)

lemma REL_SkipR:"{} \<rhd> Skip : (RELSecR R (SkipPhi R))"
(*<*)
apply (rule VDMConseq) apply (rule VDMSkip)
  apply (simp add: RELSecR_def SkipPhi_def) 
done
(*>*)

lemma REL_Skip: "RHL R Skip Skip R (SkipPhi R)"
(*<*)
apply (simp add: RHL_def, rule)
  apply (rule REL_SkipL)
  apply (rule REL_SkipR)
done
(*>*)

constdefs AssignPhi::"Var \<Rightarrow> Expr \<Rightarrow> REL \<Rightarrow> TT"
"AssignPhi x' e' S \<equiv> \<lambda> (s,t) . (s, update t x' (evalE e' t)) : S"

lemma REL_AssignL:"
  \<lbrakk>R = {(s, s'). (update s x (evalE e s), update s' x' (evalE e' s')) \<in> S}\<rbrakk>
  \<Longrightarrow> {} \<rhd> Assign x e : RELSecL R (AssignPhi x' e' S)"
(*<*)
apply (rule VDMConseq) apply (rule VDMAssign)
  apply (simp add: RELSecL_def AssignPhi_def) 
done
(*>*)

lemma REL_AssignR:
  "{} \<rhd> Assign x' e' : RELSecR S (AssignPhi x' e' S)"
(*<*)
apply (rule VDMConseq) apply (rule VDMAssign)
  apply (simp add: RELSecR_def AssignPhi_def) 
done
(*>*)

lemma  REL_Assign: 
"\<lbrakk>R = {(s,s') . (update s x (evalE e s), update s' x' (evalE e' s')) : S};
 phi = AssignPhi x' e' S\<rbrakk> \<Longrightarrow>
 RHL R (Assign x e) (Assign x' e') S phi"
(*<*)
apply (simp add: RHL_def, rule)
  apply (rule REL_AssignL) apply simp 
  apply (rule REL_AssignR) 
done
(*>*)

constdefs AssignPhi_Alt::"Var \<Rightarrow> Expr \<Rightarrow> REL \<Rightarrow> TT"
"AssignPhi_Alt x e R \<equiv> \<lambda> (s,t) . \<exists> u .s = update u x (evalE e u) \<and> (u,t) : R"

lemma REL_AssignL_Alt:
 " \<lbrakk>R = {(s, s'). (update s x (evalE e s), update s' x' (evalE e' s')) \<in> S}\<rbrakk>
  \<Longrightarrow> {} \<rhd> Assign x e : RELSecL R (AssignPhi_Alt x e R)"
(*<*)
apply (rule VDMConseq) apply (rule VDMAssign)
  apply (simp add: RELSecL_def AssignPhi_Alt_def) 
  apply clarsimp  apply (rule_tac x=s in exI, simp)
done 
(*>*)

lemma REL_AssignR_Alt:
  "\<lbrakk>R = {(s, s'). (update s x (evalE e s), update s' x' (evalE e' s')) \<in> S}\<rbrakk>
    \<Longrightarrow> {} \<rhd> Assign x' e' : RELSecR S (AssignPhi_Alt x e R)"
(*<*)
apply (rule VDMConseq) apply (rule VDMAssign) 
  apply (simp add: RELSecR_def AssignPhi_Alt_def) 
  apply clarsimp  
done 
(*>*)

lemma REL_Assign_Alt: 
"\<lbrakk>R = {(s,s') . (update s x (evalE e s), update s' x' (evalE e' s')) : S}\<rbrakk> \<Longrightarrow>
 RHL R (Assign x e) (Assign x' e') S (AssignPhi_Alt x e R)"
(*<*)
apply (simp add: RHL_def, rule)
  apply (rule_tac x'=x' and e'=e' in REL_AssignL_Alt) apply simp
  apply (rule REL_AssignR_Alt) apply simp
done
(*>*)

constdefs CompPhi::"REL \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"CompPhi T phi1 phi2 \<equiv> \<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> (\<forall> w . (r, w):T \<longrightarrow> phi2(s, w))"

lemma REL_CompL:"\<lbrakk>{} \<rhd> C1 : RELSecL R phi1; {} \<rhd> C2 : RELSecL T phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Comp C1 C2 : RELSecL R (CompPhi T phi1 phi2)"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
apply (simp add: RELSecL_def CompPhi_def, clarsimp) 
    apply (erule_tac x=ra in allE, clarsimp)
        apply (rule_tac x=r in exI, clarsimp) 
done
(*>*)

lemma REL_CompR: "\<lbrakk>{} \<rhd> D1 : RELSecR T phi1; {} \<rhd> D2 : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Comp D1 D2 : RELSecR S (CompPhi T phi1 phi2)"
(*<*)
  apply (rule VDMConseq) apply (erule VDMComp) apply assumption
    apply (simp add: RELSecR_def CompPhi_def) apply clarsimp 
done
(*>*)

lemma REL_Comp: 
  "\<lbrakk> RHL R C1 D1 T phi1; RHL T C2 D2 S phi2\<rbrakk> \<Longrightarrow>
     RHL R (Comp C1 C2) (Comp D1 D2) S (CompPhi T phi1 phi2)"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply rule
  apply (erule REL_CompL) apply assumption
  apply (erule REL_CompR) apply assumption
done
(*>*)

constdefs IfPhi::"BExpr \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"IfPhi b' phi1 phi2 \<equiv> \<lambda> (s, t) .  (evalB b' t \<longrightarrow> phi1(s,t)) \<and> 
                                  ((\<not> evalB b' t) \<longrightarrow> phi2(s,t))"

lemma REL_IfL:
  " \<lbrakk>{} \<rhd> C1 : RELSecL {(s, s'). (s, s') \<in> RR \<and> evalB b s \<and> evalB b' s'} phi1;
     {} \<rhd> C2 : RELSecL {(s, s'). (s, s') \<in> RR \<and> \<not> evalB b s \<and> \<not> evalB b' s'} phi2;
      R = {(s,s') . (s,s'):RR \<and> evalB b s = evalB b' s'}\<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff b C1 C2 : RELSecL R (IfPhi b' phi1 phi2)"
(*<*)
apply clarsimp
 apply (rule VDMConseq) apply(erule VDMIff) apply assumption+ apply (erule thin_rl)+
   apply (simp add: RELSecL_def IfPhi_def) 
done
(*>*)

lemma REL_IfR:
  "\<lbrakk> {} \<rhd> D1 : RELSecR S phi1; {} \<rhd> D2 : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff b' D1 D2 : RELSecR S (IfPhi b' phi1 phi2)"
(*<*)
 apply (rule VDMConseq) apply(erule VDMIff) apply assumption+ apply (erule thin_rl)+
   apply (simp add: RELSecR_def IfPhi_def) apply clarsimp  
   apply (case_tac "evalB b' s", clarsimp) apply clarsimp
done
(*>*)

lemma REL_If:
  "\<lbrakk> RHL R1 C C' S phi1; RHL R2 D D' S phi2;
     R1 = {(s,s') . (s,s'):RR \<and> evalB b s \<and> evalB b' s'};
     R2 = {(s,s') . (s,s'):RR \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')}; 
     R = {(s,s') . (s,s'):RR \<and> evalB b s = evalB b' s'}\<rbrakk> \<Longrightarrow>
     RHL R (Iff b C D) (Iff b' C' D') S (IfPhi b' phi1 phi2)"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule REL_IfL) apply assumption apply simp
  apply (erule REL_IfR) apply assumption
done
(*>*)

constdefs RHLPhiWhileOp::"BExpr \<Rightarrow> ((State \<times> State) set) \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"RHLPhiWhileOp b R phi \<equiv> 
     \<lambda> psi . (\<lambda>(s, t).
             (evalB b t \<longrightarrow> (\<exists>r. phi (r, t) \<and> 
                                 (\<forall>w. (r,w):R \<longrightarrow> psi (s, w)))) \<and> 
             (\<not> evalB b t \<longrightarrow> (s,t):R))"

text{*Since this operator is monotone, \ldots*}

lemma RHLPhiWhileOp_Monotone: "Monotone (RHLPhiWhileOp b R phi)"
(*<*)
apply (simp add: RHLPhiWhileOp_def Monotone_def) apply clarsimp
  apply (rule_tac x=r in exI, simp)
done
(*>*)

text{*we may define its fixed point, *}

constdefs RHLPhiWhile::"BExpr \<Rightarrow> ((State \<times> State) set) \<Rightarrow> TT \<Rightarrow> TT"
"RHLPhiWhile b R phi \<equiv> FIX (RHLPhiWhileOp b R phi)"

lemma REL_WhileL:
  " \<lbrakk>{} \<rhd> C : RELSecL R phi; \<forall>s s'. (s, s') \<in> R \<longrightarrow> evalB b s = evalB b' s'\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R (RHLPhiWhile b' R phi)"
(*<*)
apply (rule VDMConseq) apply (rule VDMWhile) 
  prefer 4 apply (subgoal_tac "\<forall>s t. (RELSecL R (RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi))) s t \<and> 
                                      \<not> evalB b t \<longrightarrow> (RELSecL R (RHLPhiWhile b' R phi)) s t") apply assumption
  apply clarsimp apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi)= RHLPhiWhile b' R phi", clarsimp)
                 apply (simp add: RHLPhiWhile_def) apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
  apply assumption
  apply clarsimp apply (simp add: RELSecL_def) 
      apply clarsimp  apply (simp add: RHLPhiWhileOp_def) 
  (*  apply (rule, clarsimp) apply (simp add: PhiWhileOp_def)*)
  apply clarsimp apply (simp add: RELSecL_def)
    apply clarsimp
      apply (simp (no_asm_simp) add: RHLPhiWhileOp_def)
      apply (rule_tac x=r in exI, rule) apply simp
      apply clarsimp
      apply (rotate_tac 4, erule_tac x=w in allE, clarsimp)
      apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi) = RHLPhiWhile b' R phi", clarsimp)
      apply (simp add: RHLPhiWhile_def)
      apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
done
(*>*)

lemma REL_WhileR:
  "\<lbrakk> {} \<rhd> C' : RELSecR R phi; \<forall>s s'. (s, s') \<in> R \<longrightarrow> evalB b s = evalB b' s';
     S = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'}\<rbrakk>  
    \<Longrightarrow> {} \<rhd> While b' C' : RELSecR S (RHLPhiWhile b' R phi)"
(*<*)
apply clarsimp
apply (rule VDMConseq)
  apply (rule VDMWhile) 
  prefer 4 apply (subgoal_tac "\<forall>s t. (RELSecR {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'} (RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi))) s t \<and>
                           \<not> evalB b' t \<longrightarrow> RELSecR {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'} (RHLPhiWhile b' R phi) s t") apply assumption
  apply clarsimp apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi)= RHLPhiWhile b' R phi", clarsimp)
                 apply (simp add: RHLPhiWhile_def) apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
  apply assumption
  apply clarsimp apply (simp add: RELSecR_def) apply clarsimp  apply (simp add: RHLPhiWhileOp_def)
   (* apply clarsimp apply (simp add: PhiWhileOp_def)*)
  apply clarsimp apply (simp add: RELSecR_def)
    apply clarsimp
      apply (subgoal_tac "\<exists>r. phi (r, s) \<and> (\<forall>w. (r, w) : R \<longrightarrow> (RHLPhiWhile b' R phi) (ra, w))")
      prefer 2 apply (simp add: RHLPhiWhileOp_def) 
      apply clarsimp apply (rotate_tac -3, erule thin_rl)
      apply (rotate_tac -1, erule_tac x=ra in allE, erule mp)
      apply (rotate_tac 1, erule_tac x=r in allE, erule impE) apply fast
      apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi) = RHLPhiWhile b' R phi", clarsimp)
      apply (simp add: RHLPhiWhile_def)
      apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
done
(*>*)

lemma RHL_WHILE:  
  "\<lbrakk> RHL R C C' R phi;
     \<forall> s s' . (s,s'):R \<longrightarrow> evalB b s = evalB b' s';
     S = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')}\<rbrakk> \<Longrightarrow>
   RHL R (While b C) (While b' C') S (RHLPhiWhile b' R phi)"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule REL_WhileL, assumption)
  apply (erule REL_WhileR, assumption) apply simp
done
(*>*)

lemma RELWhileLOrig:
  "\<lbrakk> {} \<rhd> C : RELSecL {(s, s'). (s, s') \<in> T \<and> evalB b s \<and> evalB b' s'} phi;
     R= {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'}\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R (RHLPhiWhile b' R  phi)"                (*<*)
apply (rule VDMConseq) apply (rule VDMWhile) 
  prefer 4 apply (subgoal_tac "\<forall>s t. (RELSecL R (RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi))) s t \<and> 
                                      \<not> evalB b t \<longrightarrow> (RELSecL R (RHLPhiWhile b' R phi)) s t") apply assumption
  apply (rule, rule, rule)
       apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi)= RHLPhiWhile b' R phi", clarsimp)
                 apply (simp add: RHLPhiWhile_def) apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
  apply assumption
  apply clarsimp apply (simp add: RELSecL_def) 
      apply clarsimp  apply (simp add: RHLPhiWhileOp_def) 
  (*  apply (rule, clarsimp) apply (simp add: PhiWhileOp_def)*)
  apply clarsimp apply (simp add: RELSecL_def)
    apply clarsimp
      apply (simp (no_asm_simp) add: RHLPhiWhileOp_def)
      apply (rule_tac x=r in exI, rule) apply simp
      apply clarsimp
      apply (rotate_tac 4, erule_tac x=w in allE, clarsimp)
      apply (subgoal_tac "RHLPhiWhileOp b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi (RHLPhiWhile b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi) = RHLPhiWhile b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi", clarsimp)
      apply (simp add: RHLPhiWhile_def)
      apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
done
(*>*)

lemma RELWhileROrig:
  "\<lbrakk> {} \<rhd> C' : RELSecR R phi;
     R= {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'}\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b'
             C' : RELSecR {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'}
                   (RHLPhiWhile b' R phi)"
(*<*)
apply (rule VDMConseq)
  apply (rule VDMWhile) 
  prefer 4 apply (subgoal_tac "\<forall>s t. (RELSecR {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'} 
                                              (RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi))) s t \<and>
                           \<not> evalB b' t \<longrightarrow> RELSecR {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'}
                                             (RHLPhiWhile b' R phi) s t") apply assumption
  apply clarsimp apply (subgoal_tac "RHLPhiWhileOp b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi (RHLPhiWhile b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi)= RHLPhiWhile b' {(s, s'). (s, s') \<in> T \<and> evalB b s = evalB b' s'} phi", clarsimp)
                 apply (simp add: RHLPhiWhile_def) apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
  apply assumption
  apply clarsimp apply (simp add: RELSecR_def) apply clarsimp  apply (simp add: RHLPhiWhileOp_def)
   (* apply clarsimp apply (simp add: PhiWhileOp_def)*)
  apply (erule thin_rl) apply (erule thin_rl) apply (unfold RELSecR_def)
    apply (rule, rule, rule, rule)
    apply (rule, rule) 
    apply (rule, rule)
      apply (subgoal_tac "\<exists>r. phi (r, s) \<and> (\<forall>w. (r, w) : R \<longrightarrow> (RHLPhiWhile b' R phi) (ra, w))")
      prefer 2 apply (simp add: RHLPhiWhileOp_def) 
      apply (erule exE, erule conjE) apply (rotate_tac -3, erule thin_rl)
      apply (rotate_tac -1, erule_tac x=ra in allE, erule mp)
      apply (rotate_tac 1, erule_tac x=r in allE, erule impE) apply fast
      apply (subgoal_tac "RHLPhiWhileOp b' R phi (RHLPhiWhile b' R phi) = RHLPhiWhile b' R phi")
           apply (erule thin_rl)  apply clarsimp
      apply (erule thin_rl)
      apply (simp add: RHLPhiWhile_def)
      apply (rule Fix_lemma) apply (rule RHLPhiWhileOp_Monotone)
done
(*>*)

lemma RHL_WHILEOrig:  
  "\<lbrakk> RHL U C C' R phi;
     U = {(s,s') . (s,s'):T \<and> evalB b s \<and> evalB b' s'};
     R = {(s,s') . (s,s'):T \<and> evalB b s = evalB b' s'};
     S = {(s,s') . (s,s'):T \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')}\<rbrakk> \<Longrightarrow>
   RHL R (While b C) (While b' C') S (RHLPhiWhile b' R phi)"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule_tac T=T in RELWhileLOrig) apply simp 
  apply (erule_tac T=T in RELWhileROrig) apply simp 
done
(*>*)
 

lemma REL_SubL:
  " \<lbrakk>{} \<rhd> C : RELSecL RR phi; R \<subseteq> RR\<rbrakk> \<Longrightarrow> {} \<rhd> C : RELSecL R phi"
(*<*)
  apply (erule VDMConseq) apply (simp add: RELSecL_def) apply fast
done
(*>*)

lemma REL_SubR: 
  "\<lbrakk>{} \<rhd> D : RELSecR SS phi; SS \<subseteq> S\<rbrakk> \<Longrightarrow> {} \<rhd> D : RELSecR S phi"
(*<*)
  apply (erule VDMConseq) apply (simp add: RELSecR_def) apply fast
done
(*>*)

lemma RHLSub:
 "\<lbrakk>RHL RR C D SS phi; R \<subseteq> RR; SS \<subseteq> S\<rbrakk> \<Longrightarrow> RHL R C D S phi"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply rule
  apply (erule REL_SubL) apply assumption
  apply (erule REL_SubR) apply assumption
done
(*>*)

constdefs FalsePhi::TT
"FalsePhi \<equiv> (\<lambda> (s,t) . False)"

text{*SecL works here for arbitrary phi.*}

lemma REL_FalseL:" {} \<rhd> C : RELSecL {} phi"
(*<*)
apply (rule VDMConseq) apply (rule VDMTrue) apply (simp add: RELSecL_def)
done
(*>*)

lemma REL_FalseR:"{} \<rhd> D : RELSecR S FalsePhi"
(*<*)
apply (rule VDMConseq) 
apply (rule VDMTrue) apply (simp add: RELSecR_def FalsePhi_def)
done
(*>*)

lemma RHLFalse: "RHL {} C D S FalsePhi"
(*<*)
apply (simp add: RHL_def)
apply rule  
  apply (rule REL_FalseL)
  apply (rule REL_FalseR)
done
(*>*)

(*
text{*The following alternative requires us to include rule
VDMOpsem.*}

lemma "RHL {} C D S (\<lambda> (s,t) . \<forall> r . t,D \<Down> r \<longrightarrow> (s,r):S) "
apply (simp add: RHL_def)
apply rule 
  apply (rule VDMConseq) apply (rule VDMTrue) apply (simp add: RELSecL_def) 
  apply (rule VDMConseq) apply (rule VDMOpsem) apply (simp add: RELSecR_def) 
oops
*)

constdefs DeadAssignLeftPhi::"REL \<Rightarrow> TT"
" DeadAssignLeftPhi S \<equiv> \<lambda> (s,t) . (s,t):S"

lemma REL_DeadAssignLeftL:
 "T = {(s, s'). (update s x (evalE e s), s') \<in> S} \<Longrightarrow>
 {} \<rhd> Assign x e : RELSecL T (DeadAssignLeftPhi S)"
(*<*)
apply (rule VDMConseq) apply (rule VDMAssign)
    apply (simp add: RELSecL_def DeadAssignLeftPhi_def) 
done 
(*>*)

lemma REL_DeadAssignLeftR:
"{} \<rhd> Skip : RELSecR S (DeadAssignLeftPhi S)"
(*<*)
  apply (rule VDMConseq) apply (rule VDMSkip)
    apply (simp add: RELSecR_def DeadAssignLeftPhi_def)
done
(*>*)

lemma RHLDeadAssignmentLeft:
"T = {(s,s') . (update s x (evalE e s),s') : S}
 \<Longrightarrow> RHL T (Assign x e) Skip S (DeadAssignLeftPhi S)"
(*<*)
apply (simp add: RHL_def)
apply rule
  apply (rule REL_DeadAssignLeftL) apply simp
  apply (rule REL_DeadAssignLeftR) 
done
(*>*)

constdefs DeadAssignRightPhi::"Var \<Rightarrow> Expr \<Rightarrow> REL \<Rightarrow> TT"
"DeadAssignRightPhi x e S \<equiv>  (\<lambda>(s, t). (s, update t x (evalE e t)) \<in> S)"

lemma REL_DeadAssignRightL:
"T = {(s, s'). (s, update s' x (evalE e s')) \<in> S} \<Longrightarrow>
    {} \<rhd> Skip : RELSecL T (DeadAssignRightPhi x e S)"
(*<*)
  apply (rule VDMConseq) apply (rule VDMSkip)
    apply (simp add: RELSecL_def DeadAssignRightPhi_def)
done 
(*>*)

lemma REL_DeadAssignRightR:
"{} \<rhd> Assign x e : RELSecR S (DeadAssignRightPhi x e S)"
(*<*)
  apply (rule VDMConseq) apply (rule VDMAssign)
    apply (simp add: RELSecR_def DeadAssignRightPhi_def) 
done
(*>*)
                
lemma RHLDeadAssignmentRight:
"T = {(s,s') . (s, update s' x (evalE e s')) : S}
 \<Longrightarrow> RHL T Skip (Assign x e) S (DeadAssignRightPhi x e S)"
(*<*)
apply (simp add: RHL_def)
apply rule
  apply (rule REL_DeadAssignRightL) apply simp
  apply (rule REL_DeadAssignRightR)
done
(*>*)

lemma "\<lbrakk>RHL T C D S1 phi1; RHL T C D S2 phi2;
 S = S1 \<inter> S2; phi = (\<lambda> (s, t) . phi1 (s, t) \<and> phi2 (s, t))\<rbrakk>
\<Longrightarrow> RHL T C D S phi"  
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule 
  apply (erule VDMInter) apply assumption
  apply rule apply rule apply (simp add: RELSecL_def) 
  apply rule apply rule apply clarsimp apply clarsimp 
  apply clarsimp
(*>*)
oops
(*Works for SecL*)

(*Neither \<and> nor \<and>  appear to work for SecR, neither for union nor for intersection of S1 S2*)

lemma "\<lbrakk>RHL T C D S1 phi1; RHL T C D S2 phi2;
 S = S1 \<union> S2; phi = (\<lambda> (s, t) . phi1 (s, t) \<or> phi2 (s, t))\<rbrakk>
\<Longrightarrow> RHL T C D S phi"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule defer 
apply (erule VDMInter) apply assumption
  apply rule apply rule apply (simp add: RELSecR_def) 
(*>*)
oops    

(*Note that we can only prove the following rule if T and S occur
identically in both hypotheses, and the conclsuion. Union or
Intesection on don't seem to work*)

constdefs AndPhi::"TT \<Rightarrow> TT \<Rightarrow> TT"
"AndPhi phi1 phi2 \<equiv> \<lambda> (s, t) . phi1 (s, t) \<and> phi2 (s, t)"

lemma REL_AndL:
 "\<lbrakk>{} \<rhd> C : RELSecL T phi1; {} \<rhd> C : RELSecL T phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecL T (AndPhi phi1 phi2)"
(*<*)
  apply (erule VDMInter) apply assumption
  apply rule apply rule apply (simp add: RELSecL_def AndPhi_def)
  apply rule apply rule apply clarsimp apply clarsimp 
  apply clarsimp
done
(*>*)

lemma REL_AndR:
  "\<lbrakk>{} \<rhd> D : RELSecR S phi1; {} \<rhd> D : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecR S (AndPhi phi1 phi2)"
(*<*)
apply (rule VDMConseq)
  apply (erule VDMInter) apply assumption apply simp 
  apply (simp add: RELSecR_def AndPhi_def)
done
(*>*)

lemma RHLPhiAnd:
 "\<lbrakk>RHL T C D S phi1; RHL T C D S phi2\<rbrakk>
\<Longrightarrow> RHL T C D S (AndPhi phi1 phi2)"  
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule 
  apply (erule REL_AndL) apply assumption
  apply (erule REL_AndR) apply assumption
done
(*>*)

lemma REL_UnionL:
" \<lbrakk>G \<rhd> C : RELSecL R1 phi1; G \<rhd> C : RELSecL R2 phi2; 
    R = R1 \<union> R2; phi = (\<lambda> (s,t) . phi1(s,t) \<or> phi2(s,t)) \<rbrakk>
    \<Longrightarrow> G \<rhd> C : RELSecL R phi"
apply clarsimp
  apply (rule VDMConseq) apply (erule VDMInter,assumption) apply simp 
  apply (erule thin_rl)+ apply (simp add: RELSecL_def) 
done

lemma REL_UnionR:
 "\<lbrakk> G \<rhd> D : RELSecR S phi1; G \<rhd> D : RELSecR S phi2; 
    phi = (\<lambda> (s,t) . phi1(s,t) \<or> phi2(s,t))\<rbrakk>
  \<Longrightarrow> G \<rhd> D : RELSecR S phi"
apply clarsimp 
apply (rule VDMConseq) apply (erule VDMInter,assumption) apply simp 
  apply (erule thin_rl)+ apply (simp add: RELSecR_def) 
done

lemma RHLUnion:
  "\<lbrakk> RHL R1 C D S phi1;
     RHL R2 C D S phi2;
     R = R1 \<union> R2; phi = (\<lambda> (s,t) . phi1(s,t) \<or> phi2(s,t)) \<rbrakk> \<Longrightarrow>
     RHL R C D S phi"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule REL_UnionL) apply assumption apply simp apply simp
  apply (erule REL_UnionR) apply assumption apply simp
done
(*>*)

text{*We use the same trick that establishes RHLPhiAnd, i.e. the
application of an itermediate VDMConseq-VDMInter pair, in the proof of
the CommonBranch rules.*}

constdefs CBRLeftPhi::"TT \<Rightarrow> TT \<Rightarrow> TT"
"CBRLeftPhi phi1 phi2 \<equiv> \<lambda> (s,t) . phi1 (s,t) \<or> phi2 (s,t)"

lemma REL_CBRLeftL:
 "\<lbrakk> {} \<rhd> C1 : RELSecL {(s, s'). (s, s') \<in> R \<and> evalB B s} phi1;
    {} \<rhd> C2 : RELSecL {(s, s'). (s, s') \<in> R \<and> \<not> evalB B s} phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff B C1 C2 : RELSecL R (CBRLeftPhi phi1 phi2)"
(*<*)
apply (rule VDMConseq) apply (rule VDMIff) apply assumption apply assumption
    apply (erule thin_rl)+ 
    apply (simp add: RELSecL_def CBRLeftPhi_def, clarsimp)
    apply (case_tac "evalB B s", clarsimp) apply clarsimp
done
(*>*)

lemma REL_CBRLeftR:
 "\<lbrakk> {} \<rhd> D : RELSecR S phi1; {} \<rhd> D : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecR S (CBRLeftPhi phi1 phi2)"
(*<*)
apply (rule VDMConseq) apply (erule VDMInter) apply assumption apply simp 
apply (erule thin_rl)+ 
apply (simp add: RELSecR_def CBRLeftPhi_def) 
done
(*>*)

lemma RHLCommonBranchLeft:
 "\<lbrakk> RHL T1 C1 D S phi1;
    RHL T2 C2 D S phi2;
    T1 = {(s, s') . (s,s'):T \<and> evalB B s};
    T2 = {(s, s') . (s,s'):T \<and> \<not> evalB B s}\<rbrakk> 
  \<Longrightarrow> RHL T (Iff B C1 C2) D S (CBRLeftPhi phi1 phi2)"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule REL_CBRLeftL, assumption)
  apply (erule REL_CBRLeftR, assumption)
done
(*>*)

text{*Note that CommonBranchRight uses a different construction of phi
that CommonBranchLeft. This is not surprising since (the
interpretation of) our judgement form is not symmetric!*}

constdefs CBRRightPhi::"BExpr \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"CBRRightPhi b phi1 phi2 \<equiv>
  \<lambda> (s,t) . (evalB b t \<and> phi1 (s,t)) \<or>
            ((\<not> evalB b t) \<and> phi2 (s,t))"

lemma REL_CBRRightL:
  "\<lbrakk> {} \<rhd> C : RELSecL {(s, s'). (s, s') \<in> T \<and> evalB B s'} phi1;
     {} \<rhd> C : RELSecL {(s, s'). (s, s') \<in> T \<and> \<not> evalB B s'} phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecL T (CBRRightPhi B phi1 phi2)"
(*<*)
  apply (rule VDMConseq) apply (erule VDMInter) apply assumption apply simp 
    apply (erule thin_rl)+ 
    apply (simp add: RELSecL_def CBRRightPhi_def) apply clarsimp
    apply (case_tac "evalB B r") apply clarsimp apply clarsimp 
done
(*>*)

lemma REL_CBRRightR:
  "\<lbrakk> {} \<rhd> D1 : RELSecR S phi1; {} \<rhd> D2 : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff B D1 D2 : RELSecR S (CBRRightPhi B phi1 phi2)"
(*<*)
apply (rule VDMConseq) apply (erule VDMIff) apply assumption 
    apply (erule thin_rl)+ 
    apply (simp add: RELSecR_def CBRRightPhi_def) 
done
(*>*)

lemma RHLCommonBranchRight:
 "\<lbrakk> RHL T1 C D1 S phi1;
    RHL T2 C D2 S phi2;
    T1 = {(s, s') . (s,s'):T \<and> evalB B s'};
    T2 = {(s, s') . (s,s'):T \<and> \<not> evalB B s'}\<rbrakk> 
  \<Longrightarrow> RHL T C (Iff B D1 D2) S (CBRRightPhi B phi1 phi2)"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (erule REL_CBRRightL, assumption)
  apply (erule REL_CBRRightR, assumption)
done
(*>*)

lemma REL_DeadWhileLeftL:
  "\<lbrakk>R = {(s, s'). (s, s') \<in> S \<and> \<not> evalB b s}; phi = (\<lambda> (s,t) . (s,t): R )\<rbrakk>
       \<Longrightarrow> {} \<rhd> While b C : RELSecL R phi"
(*<*)
  apply (rule VDMConseq) apply (rule_tac I="(\<lambda> x y . (x,y) :R)" in VDMWhile_Hoare) 
   apply (rule VDMConseq) apply (rule VDMTrue) apply clarsimp apply (simp add: RELSecL_def) apply clarsimp 
done
(*>*)

text{*The following is just a simple application of REL-SkipR*}

lemma REL_DeadWhileLeftR:
  "R = {(s,s') . (s,s'):S \<and> \<not> evalB b s} \<Longrightarrow> {} \<rhd> Skip : RELSecR R (\<lambda>(s, t). (s, t) \<in> S \<and> \<not> evalB b s)"
(*<*)
  apply (rule VDMConseq)
  apply (rule_tac R=R in REL_SkipR) apply (simp add: SkipPhi_def)  
done 
(*>*)

lemma RHLDeadWhileLeft:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s}; phi = (\<lambda> (s,t) . (s,t): R )\<rbrakk> 
       \<Longrightarrow> RHL R (While b C) Skip R phi"
(*<*)
apply (simp add: RHL_def)
apply rule 
  apply (rule_tac S=S in REL_DeadWhileLeftL) apply simp apply simp 
  apply (rule REL_DeadWhileLeftR[simplified]) apply simp
done
(*>*)

text{*The following is just a simple application of REL-SkipL*}

lemma REL_DeadWhileRightL:
  "\<lbrakk>R = {(s, s'). (s, s') \<in> S \<and> \<not> evalB b s'}; phi = (\<lambda>(s, t). (s, t) \<in> S \<and> \<not> evalB b t)\<rbrakk>
    \<Longrightarrow> {} \<rhd> Skip : RELSecL R phi"
(*<*)
  apply (rule VDMConseq)
  apply (rule_tac R=R in REL_SkipL) apply (simp add: SkipPhi_def)  
done 
(*>*)

lemma REL_DeadWhileRightR:
  " \<lbrakk>R = {(s, s'). (s, s') \<in> S \<and> \<not> evalB b s'}; phi = (\<lambda>(s, t). (s, t) \<in> S \<and> \<not> evalB b t)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecR R phi"
(*<*)
  apply (rule VDMConseq) apply (rule_tac I="(\<lambda> x y . (y,x) :R)" in VDMWhile_Hoare) 
   apply (rule VDMConseq) apply (rule VDMTrue) apply clarsimp apply (simp add: RELSecR_def) apply clarsimp 
done
(*>*)

lemma RHLDeadWhileRight:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s'}; phi = (\<lambda> (s,t) . (s,t): R) \<rbrakk> 
       \<Longrightarrow> RHL R Skip (While b C) R phi"
(*<*)
apply (simp add: RHL_def)
apply rule 
  apply (rule_tac S=S and b=b in REL_DeadWhileRightL[simplified]) apply simp apply simp
  apply (rule_tac S=S and b=b in REL_DeadWhileRightR) apply simp apply simp 
done
(*>*)

lemma DeadLeftL:
  "\<lbrakk>{} \<rhd> C : A; \<forall>s s' t. A s t \<longrightarrow> (s, s') \<in> R \<longrightarrow> (t, s') \<in> S;
    phi = (\<lambda> (s, t) . (s,t):S)\<rbrakk> \<Longrightarrow> {} \<rhd> C : RELSecL R phi"
(*<*)
apply (erule VDMConseq)
apply (simp add: RELSecL_def) 
done
(*>*)

lemma DeadLeftR:
  "\<lbrakk>{} \<rhd> C : A; \<forall>s s' t. A s t \<longrightarrow> (s, s') \<in> R \<longrightarrow> (t, s') \<in> S; phi = (\<lambda>(s, t). (s, t) \<in> S)\<rbrakk>
    \<Longrightarrow> {} \<rhd> Skip : RELSecR S phi"
(*<*)
apply (rule_tac A="RELSecR S (SkipPhi S)" in VDMConseq)
apply (rule REL_SkipR)
apply (simp add: RELSecR_def)
apply (simp add:SkipPhi_def)
done
(*>*)

lemma DeadLeft: 
  "\<lbrakk> {} \<rhd> C : A; \<forall> s s' t . A s t \<longrightarrow> (s,s'):R \<longrightarrow> (t,s'):S;
     phi = (\<lambda> (s, t) . (s,t):S)\<rbrakk>
  \<Longrightarrow> RHL R C Skip S phi"
(*<*)
apply (simp add: RHL_def)
apply rule 
  apply (erule_tac S=S in DeadLeftL) apply assumption apply simp 
  apply (erule_tac R=R in DeadLeftR) apply assumption apply simp
done
(*>*)

lemma RHLDeadWhileLeft_derivable:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s}; phi = (\<lambda> (s,t) . (s,t): R )\<rbrakk> 
       \<Longrightarrow> RHL R (While b C) Skip R phi"
(*<*)
apply (rule DeadLeft) 
apply (rule_tac A="\<lambda> s t . \<forall> s' . (s,s'):R \<longrightarrow> (t,s'):R" in VDMWhile) apply (rule VDMTrue)
  apply simp apply clarsimp apply clarsimp 
  apply simp
done
(*>*)

lemma DeadRightL:
  "\<lbrakk>{} \<rhd> C : A; \<forall>s s' t'. A s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t') \<in> S;
    phi = (\<lambda> (s,t). (s,t):R)\<rbrakk> \<Longrightarrow> {} \<rhd> Skip : RELSecL R phi"
(*<*)
apply (rule VDMConseq)
apply (rule VDMSkip)
apply (simp add: RELSecL_def)
done
(*>*)

lemma DeadRightR:
  " \<lbrakk>{} \<rhd> C : A; \<forall>s s' t'. A s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t') \<in> S; phi = (\<lambda>(s, t). (s, t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecR S phi"
(*<*)
apply (erule VDMConseq)
apply (simp add: RELSecR_def)
done
(*>*)

lemma DeadRight: 
  "\<lbrakk> {} \<rhd> C : A; \<forall> s s' t' . A s' t' \<longrightarrow> (s,s'):R \<longrightarrow> (s,t'):S;
    phi = (\<lambda> (s,t). (s,t):R)\<rbrakk>
  \<Longrightarrow> RHL R Skip C S phi"
(*<*)
apply (simp add: RHL_def)
apply rule 
  apply (erule_tac S=S in DeadRightL) apply assumption apply simp 
  apply (erule_tac R=R in DeadRightR) apply assumption apply simp
done
(*>*)

lemma RHLDeadWhileRight_derivable:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s'}; phi = (\<lambda> (s,t) . (s,t): R) \<rbrakk> 
       \<Longrightarrow> RHL R Skip (While b C) R phi"
(*<*)
apply (rule DeadRight)
apply (rule_tac A="\<lambda> s' t' . \<forall> s . (s,s'):R \<longrightarrow> (s,t'):R" in VDMWhile) apply (rule VDMTrue)
  apply simp apply clarsimp apply clarsimp 
  apply simp
done
(*>*)

lemma RHLDeadAssignmentLeft_derivable:
"T = {(s,s') . (update s x (evalE e s),s') : S}
 \<Longrightarrow> RHL T (Assign x e) Skip S (DeadAssignLeftPhi S)"
(*<*)
apply (rule DeadLeft)
apply (rule VDMAssign)
apply clarsimp
apply (simp add: DeadAssignLeftPhi_def)
done
(*>*)
           
lemma RHLDeadAssignmentRight_derivable:
"T = {(s,s') . (s, update s' x (evalE e s')) : S}
 \<Longrightarrow> RHL T Skip (Assign x e) S (DeadAssignRightPhi x e S)"
(*<*)
apply (rule DeadRight)
apply (rule VDMAssign)
apply clarsimp 
apply (simp add: DeadAssignRightPhi_def)
done
(*>*)

lemma UniLeftL:
  " \<lbrakk>{} \<rhd> C : A; {} \<rhd> D : RELSecL T phi; \<forall>s s' t. A s t \<longrightarrow> (s, s') \<in> R \<longrightarrow> (t, s') \<in> T\<rbrakk>
    \<Longrightarrow> {} \<rhd> Comp C D : RELSecL R phi"
(*<*)
apply (rule VDMConseq)
apply (erule VDMComp) apply assumption
apply (simp add: RELSecL_def) apply clarsimp
done
(*>*)

lemma UniLeft: 
  "\<lbrakk> {} \<rhd> C : A; RHL T D C' S phi;
      \<forall> s s' t . A s t \<longrightarrow> (s,s'):R \<longrightarrow> (t,s'):T\<rbrakk>
  \<Longrightarrow> RHL R (Comp C D) C' S phi"
(*<*)
apply (unfold RHL_def) apply (erule conjE)
apply rule
  apply (erule_tac A=A and T=T in UniLeftL) apply assumption apply assumption 
apply assumption
done
(*>*)

lemma LeftUniL:
  "\<lbrakk>{} \<rhd> D : A; {} \<rhd> C : RELSecL R phi; 
    \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> T \<longrightarrow> (t, s') \<in> S; 
    psi = (\<lambda>(s, s'). \<exists>r. A r s \<and> phi (r, s'))\<rbrakk>
  \<Longrightarrow> {} \<rhd> Comp C D : RELSecL R psi"
(*<*)
  apply (rule VDMConseq) apply (erule VDMComp) apply assumption apply clarsimp
     apply (simp add: RELSecL_def) apply clarsimp
     apply (erule_tac x=ra in allE, erule impE, assumption) apply (rule_tac x=r in exI, simp)
done
(*>*)

lemma LeftUniR:
  "\<lbrakk>{} \<rhd> D : A;   {} \<rhd> C' : RELSecR T phi;
    \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> T \<longrightarrow> (t, s') \<in> S;
    psi = (\<lambda>(s, s'). \<exists>r. A r s \<and> phi (r, s'))\<rbrakk>
   \<Longrightarrow> {} \<rhd> C' : RELSecR S psi"
(*<*)
  apply (erule VDMConseq) apply (simp add: RELSecR_def) apply clarsimp
done
(*>*)

lemma LeftUni: 
  "\<lbrakk> {} \<rhd> D : A; RHL R C C' T phi;
      \<forall> s t s'. A s t \<longrightarrow> (s,s'):T \<longrightarrow> (t,s'):S;
      psi = (\<lambda> (s,s') . \<exists> r . A r s \<and> phi (r,s'))\<rbrakk>
  \<Longrightarrow> RHL R (Comp C D) C' S psi"
(*<*)
apply (unfold RHL_def) apply (erule conjE)
apply rule
  apply (erule_tac A=A and phi=phi and T=T and S=S in LeftUniL) apply assumption+
  apply (erule_tac D=D and A=A and phi=phi and T=T in LeftUniR) apply assumption+
done
(*>*)

lemma UniRightL:
  "\<lbrakk> {} \<rhd> C' : A';  {} \<rhd> C : RELSecL R phi;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t') \<in> R; psi = (\<lambda>(s, s'). \<forall>t'. A' s' t' \<longrightarrow> phi (s, t'))\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecL R psi"
(*<*)
  apply (erule VDMConseq) apply (simp add: RELSecL_def) 
done
(*>*)

lemma UniRightR:
  "\<lbrakk>{} \<rhd> C' : A'; {} \<rhd> D' : RELSecR S phi;
    \<forall> s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t') \<in> R;
    psi = (\<lambda> (s,s') . \<forall> t' . A' s' t' \<longrightarrow> phi (s,t')) \<rbrakk>
    \<Longrightarrow> {} \<rhd> Comp C' D' : RELSecR S psi"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
apply (simp add: RELSecR_def) apply clarsimp 
done
(*>*)

lemma UniRight: 
  "\<lbrakk> {} \<rhd> C' : A'; RHL R C D' S phi;
     \<forall> s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t') \<in> R;
      psi = (\<lambda> (s,s') . \<forall> t' . A' s' t' \<longrightarrow> phi (s,t'))\<rbrakk>
  \<Longrightarrow> RHL R C (Comp C' D') S psi"
(*<*)
apply (unfold RHL_def) apply (erule conjE)
apply rule
  apply (erule_tac C'=C' and A'=A' and phi=phi in UniRightL) 
    apply assumption apply assumption apply assumption 
  apply (erule_tac A'=A' and phi=phi and R=R in UniRightR)
    apply assumption apply assumption apply assumption 
done
(*>*)

lemma RightUniL:
  "\<lbrakk>{} \<rhd> D' : A'; {} \<rhd> C : RELSecL R phi;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> (s, t') \<in> S\<rbrakk>
    \<Longrightarrow> {} \<rhd> C : RELSecL R phi"
(*<*)
  apply (erule VDMConseq) apply clarsimp
done
(*>*)

lemma RightUniR:
  "\<lbrakk> {} \<rhd> D' : A';  {} \<rhd> C' : RELSecR T phi;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> (s, t') \<in> S\<rbrakk>
    \<Longrightarrow> {} \<rhd> Comp C' D' : RELSecR S phi"
(*<*)
  apply (rule VDMConseq)
    apply (erule VDMComp) apply assumption
    apply (simp add: RELSecR_def) apply clarsimp 
done
(*>*)

lemma RightUni: 
  "\<lbrakk> {} \<rhd> D' : A'; RHL R C C' T phi;
     \<forall> s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> (s, t') \<in> S\<rbrakk>
  \<Longrightarrow> RHL R C (Comp C' D') S phi"
(*<*)
apply (unfold RHL_def) apply (erule conjE)
apply rule
  apply (erule_tac A'=A' and D'=D' and T=T and S=S in RightUniL) apply assumption+
  apply (erule_tac A'=A' and T=T  in RightUniR) apply assumption+
done
(*>*)

constdefs IfDiffPhi::"BExpr \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"IfDiffPhi b' phi1 phi2 \<equiv>
  \<lambda> (s, t) .  (evalB b' t \<longrightarrow> phi2(s,t)) \<and> 
              ((\<not> evalB b' t) \<longrightarrow> phi1(s,t))"

lemma REL_IfDiffL:
 " \<lbrakk> {} \<rhd> C : RELSecL R1 phi1;
     {} \<rhd> D : RELSecL R2 phi2;
     R = {(s, s'). (s, s') \<in> RR \<and> evalB b s = (\<not> evalB b' s')};
     R1 = {(s,s') . (s,s'):RR \<and> evalB b s \<and> (\<not> evalB b' s')};
     R2 = {(s,s') . (s,s'):RR \<and> (\<not> evalB b s) \<and> evalB b' s'} \<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff b C
             D : RELSecL R (IfDiffPhi b' phi1 phi2)"
(*<*)
apply clarsimp
apply (rule VDMConseq) apply(erule VDMIff) apply assumption
    apply (erule thin_rl)+
    apply (simp add: RELSecL_def IfDiffPhi_def) 
done
(*>*)

lemma REL_IfDiffR:
  "\<lbrakk> {} \<rhd> D' : RELSecR S phi1; {} \<rhd> C' : RELSecR S phi2\<rbrakk>
    \<Longrightarrow> {} \<rhd> Iff b' C' D' : RELSecR S (IfDiffPhi b' phi1 phi2)"
(*<*)
apply (rule VDMConseq) apply(erule VDMIff) apply assumption apply (erule thin_rl)+
   apply (simp add: RELSecR_def IfDiffPhi_def) apply clarsimp  
   apply (case_tac "evalB b' s", clarsimp) apply clarsimp
done
(*>*)

lemma RHLIfDiff:
  "\<lbrakk> RHL R3 C D' S phi1; RHL R4 D C' S phi2;
     R3 = {(s,s') . (s,s'):RR \<and> evalB b s \<and> (\<not> evalB b' s')};
     R4 = {(s,s') . (s,s'):RR \<and> (\<not> evalB b s) \<and> evalB b' s'};
     R = {(s,s') . (s,s'):RR \<and> \<not> evalB b s = evalB b' s'} 
  \<rbrakk> \<Longrightarrow> RHL R (Iff b C D) (Iff b' C' D') S (IfDiffPhi b' phi1 phi2)"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply rule 
  apply (erule_tac RR=RR in REL_IfDiffL) apply assumption apply simp apply simp apply simp 
  apply (erule REL_IfDiffR) apply assumption 
done
(*>*)

lemma RHLFullIf:
  "\<lbrakk>  R1 = {(s,s') . (s,s'):R \<and> evalB b s \<and> evalB b' s'};
      R2 = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')};
      RHL R1 C C' S phi1;
      RHL R2 D D' S phi2;
      R3 = {(s,s') . (s,s'):R \<and> evalB b s \<and> (\<not> evalB b' s')};
      R4 = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> evalB b' s'};
      RHL R3 C D' S phi3;
      RHL R4 D C' S phi4;
      phi = (\<lambda>(s, t).
             (\<lambda>(s, t). (evalB b' t \<longrightarrow> phi1 (s, t)) \<and> (\<not> evalB b' t \<longrightarrow> phi2 (s, t))) (s, t) \<or>
             (\<lambda>(s, t). (evalB b' t \<longrightarrow> phi4 (s, t)) \<and> (\<not> evalB b' t \<longrightarrow> phi3 (s, t))) (s, t)) \<rbrakk> \<Longrightarrow>
     RHL R (Iff b C D) (Iff b' C' D') S phi"
(*<*)
apply (rule RHLUnion)
  apply (erule_tac RR=R in REL_If) apply assumption apply assumption apply assumption apply simp
  apply (rule_tac RR=R in RHLIfDiff) apply assumption apply assumption apply assumption apply assumption apply simp 
apply (erule thin_rl)+ apply fast 
apply (simp add: IfPhi_def IfDiffPhi_def)
done
(*>*)

lemma RHLFullIf_simplified:
  "\<lbrakk>  R1 = {(s,s') . (s,s'):R \<and> evalB b s \<and> evalB b' s'};
      R2 = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')};
      RHL R1 C C' S phi1;
      RHL R2 D D' S phi2;
      R3 = {(s,s') . (s,s'):R \<and> evalB b s \<and> (\<not> evalB b' s')};
      R4 = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> evalB b' s'};
      RHL R3 C D' S phi3;
      RHL R4 D C' S phi4;
      phi = 
     (\<lambda>(s, t).
           (((evalB b' t \<longrightarrow> phi1 (s, t)) \<and> ((\<not> evalB b' t) \<longrightarrow> phi2 (s, t))) \<or>
            ((evalB b' t \<longrightarrow> phi4 (s, t)) \<and> ((\<not> evalB b' t) \<longrightarrow> phi3 (s, t)))))
    \<rbrakk> \<Longrightarrow>
     RHL R (Iff b C D) (Iff b' C' D') S phi"
(*<*)
apply simp 
apply (rule RHLFullIf) prefer 3 apply assumption prefer 3 apply assumption 
      prefer 5 apply assumption prefer 5 apply assumption
  apply simp apply simp apply simp apply simp apply simp
done
(*>*)

lemma WhileL1:
  "\<lbrakk>RHL T Skip D U phi; R = {(s,s'). (s,s'):T \<and> \<not> evalB b s}; 
     phi = (\<lambda>(s, t). (s, t) \<in> R);
     S = {(t,t'). (t,t'):U \<and> \<not> evalB b t}\<rbrakk>
   \<Longrightarrow> RHL R (While b C) D S phi"
(*<*)
apply (simp add: RHL_def, clarsimp)
apply rule
  apply (rule_tac S=T in REL_DeadWhileLeftL) apply simp
  apply simp
apply (erule VDMConseq) apply clarsimp
  apply (simp add: RELSecR_def)
done
(*>*)

lemma RHL_WHILETriv:  
  "\<lbrakk> {} \<rhd> While b C : A; {} \<rhd> While b' C' : A'; 
     R = {(s, s'). \<forall>t t'. A s t \<longrightarrow> A' s' t' \<longrightarrow> (t, t') \<in> S};
     phi = (\<lambda>(s, t). \<exists>u. A u s \<and> (u, t) \<in> R) \<rbrakk>
  \<Longrightarrow> RHL R (While b C) (While b' C') S phi"
(*<*)
apply (rule REL_Unary) 
 apply assumption apply assumption apply assumption apply assumption
done
(*>*)

lemma "\<lbrakk>{} \<rhd> While b C :A;  phi = (\<lambda>(s, t). \<exists>u. A u s \<and> (u, t) \<in> R);
         R = {(s, s'). \<forall>t t'. A s t \<longrightarrow> s' = t' \<longrightarrow> (t, t') \<in> S}\<rbrakk>  \<Longrightarrow>
      RHL R (While b C) Skip S phi"
(*<*)
apply (rule REL_Unary) apply assumption apply (rule VDMSkip) apply assumption
  apply simp
done
(*>*)

lemma "\<lbrakk>{} \<rhd> While b C :A; phi = (\<lambda>(s, t). \<exists>u. s = u \<and> (u, t) \<in> R);
        R = {(s, s'). \<forall>t t'. s = t \<longrightarrow> A s' t' \<longrightarrow> (t, t') \<in> S}\<rbrakk>  \<Longrightarrow>
      RHL R Skip (While b C) S phi"
(*<*)
apply (rule REL_Unary) apply (rule VDMSkip) apply assumption apply assumption 
  apply simp
done
(*>*)

lemma "\<lbrakk>{} \<rhd> While b C :A;  {} \<rhd> D : B; phi = (\<lambda>(s, t). \<exists>u. A u s \<and> (u, t) \<in> R);
         R = {(s, s'). \<forall>t t'. A s t \<longrightarrow> B s' t' \<longrightarrow> (t, t') \<in> S}\<rbrakk>  \<Longrightarrow>
      RHL R (While b C) D S phi"
(*<*)
apply (rule REL_Unary) apply assumption apply assumption apply assumption apply assumption
done
(*>*)

lemma "\<lbrakk>{} \<rhd> While b C :A; {} \<rhd> D : B; phi = (\<lambda>(s, t). \<exists>u. B u s \<and> (u, t) \<in> R);
        R = {(s, s'). \<forall>t t'. B s t \<longrightarrow> A s' t' \<longrightarrow> (t, t') \<in> S}\<rbrakk>  \<Longrightarrow>
      RHL R D (While b C) S phi"
(*<*)
apply (rule REL_Unary) apply assumption apply assumption apply assumption apply assumption 
done
(*>*)

lemma REL_WhileLeftL:
  " \<lbrakk>{} \<rhd> C : A; \<forall>s t s'. A s t \<longrightarrow> evalB b s \<longrightarrow> ((s, s') \<in> R \<longrightarrow> (t, s') \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (t, s') \<in> T); {} \<rhd> D : B;
     R = {(s, s'). \<forall>t'. B s' t' \<longrightarrow> (s, t') \<in> T}; S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s};
     {} \<rhd> C : RELSecL R phi;
     psi = (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) \<or> 
                     ((s, t) \<in> R \<and> \<not> evalB b s))\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R psi"
(*<*)
apply (rule REL_UnionL)
prefer 3 apply (subgoal_tac "R = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s} \<union> {(s, s'). (s, s') \<in> R \<and> evalB b s}")
           apply assumption apply (erule thin_rl)+ apply rule apply fastsimp apply fastsimp
prefer 3 apply (subgoal_tac "psi = (\<lambda>(s, t). (\<lambda>(s, t). (\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) (s, t) \<or> 
                                             (\<lambda>(s,t) . (s, t) \<in> R \<and> \<not> evalB b s) (s,t))", assumption)
         apply simp
(*Case \<not> evalB b s*)
  apply (rotate_tac -1, erule thin_rl)
  apply (erule thin_rl, erule thin_rl)
  apply (rule_tac S=R in REL_DeadWhileLeftL) apply simp apply simp
(*Case evalB b s*)
apply (rotate_tac -1, erule thin_rl)
   apply (rule VDMConseq)
    apply (erule_tac A="\<lambda> x y . \<forall> s' . (x,s'):R \<longrightarrow> (y,s'):R" in VDMWhile) 
      apply clarsimp 
      apply clarsimp apply fast
  apply (simp add: RELSecL_def) apply clarsimp
done 
(*>*)

lemma REL_WhileLeftR:
  "\<lbrakk>{} \<rhd> C : A; \<forall>s t s'. A s t \<longrightarrow> evalB b s \<longrightarrow> ((s, s') \<in> R \<longrightarrow> (t, s') \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (t, s') \<in> T); {} \<rhd> D : B;
     R = {(s, s'). \<forall>t'. B s' t' \<longrightarrow> (s, t') \<in> T}; S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s};
     psi = (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) \<or> 
                     ((s, t) \<in> R \<and> \<not> evalB b s));
     {} \<rhd> D : RELSecR T phi\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecR S psi"
(*<*)
  apply (rule VDMConseq) apply (erule VDMInter) apply assumption apply simp
   apply (simp add: RELSecR_def) 
done
(*>*)

lemma REL_WhileLeft:
  "\<lbrakk> {} \<rhd> C : A;
         \<forall> s t s' . A s t \<longrightarrow> evalB b s \<longrightarrow>
                    (((s,s'): R \<longrightarrow> (t,s'):R) \<and>
                     ((s,s'): T \<longrightarrow> (t,s'):T));
          {} \<rhd> D : B;
         RHL R C D T phi;
         R = {(s,s') . \<forall> t' . B s' t' \<longrightarrow> (s,t') : T };
         S = {(s,s'). (s,s'): T \<and> (\<not> evalB b s)};
     psi = (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) \<or> 
                     ((s, t) \<in> R \<and> \<not> evalB b s))\<rbrakk>
       \<Longrightarrow> RHL R (While b C) D S psi"
(*<*)
apply (unfold RHL_def) apply (erule conjE)+ 
apply rule 
  apply (erule REL_WhileLeftL) apply assumption+
  apply (erule REL_WhileLeftR) apply assumption+
done
(*>*)

(*<*)
lemma " (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) \<or> 
                     ((s, t) \<in> R \<and> \<not> evalB b s))
      =  (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<or> (s, t) \<in> R) \<and> \<not> evalB b s)"
apply rule
apply rule apply clarsimp apply fastsimp
apply fastsimp
done

lemma "\<lbrakk> {} \<rhd> C : A;
         \<forall> s t s' . A s t \<longrightarrow> evalB b s \<longrightarrow>
                    (((s,s'): R \<longrightarrow> (t,s'):R) \<and>
                     ((s,s'): T \<longrightarrow> (t,s'):T));
          {} \<rhd> D : B;
         RHL R C D T phi;
         R = {(s,s') . \<forall> t' . B s' t' \<longrightarrow> (s,t') : T };
         S = {(s,s'). (s,s'): T \<and> (\<not> evalB b s)};
     psi = (\<lambda>(s, t). ((\<forall>t'. B t t' \<longrightarrow> (s, t') \<in> T) \<and> \<not> evalB b s) \<or> 
                     ((s, t) \<in> R \<and> \<not> evalB b s))\<rbrakk>
       \<Longrightarrow> BINsecure (Inverse R) D (While b C) (Inverse S)"
apply (subgoal_tac "BINsecure R (While b C) D S")
  apply (insert BINsecure_symmetric [of "Inverse R" D "While b C" "Inverse S"])   
  apply clarsimp apply (simp add: Inverse_def)
apply (rotate_tac -1, erule thin_rl)
apply (rule RHL_valid_implies_BIN_secure)
apply (erule REL_WhileLeft) apply assumption+
done
(*>*)

lemma REL_WhileRightL:
  "\<lbrakk>{} \<rhd> C : A; {} \<rhd> D : B;
    \<forall>s t s'. A s t \<longrightarrow> evalB b s \<longrightarrow> ((s', s) \<in> R \<longrightarrow> (s',t) \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (t, s') \<in> T); 
     R = {(s', s). \<forall>t'. B s' t' \<longrightarrow> (s, t') \<in> T}; S = {(s', s). (s, s') \<in> T \<and> \<not> evalB b s};
     psi = (\<lambda>(x, y). \<exists>s. B s x \<and> (s, y) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecL R psi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecL_def) apply clarsimp
  apply (rule_tac x=s in exI, simp)
done
(*>*)

lemma REL_WhileRightR:
   "\<lbrakk>{} \<rhd> C : A; {} \<rhd> D : B;
    \<forall>s t s'. A s t \<longrightarrow> evalB b s \<longrightarrow> ((s', s) \<in> R \<longrightarrow> (s',t) \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (t, s') \<in> T); 
     R = {(s', s). \<forall>t'. B s' t' \<longrightarrow> (s, t') \<in> T}; S = {(s', s). (s, s') \<in> T \<and> \<not> evalB b s};
     psi = (\<lambda>(x, y). \<exists>s. B s x \<and> (s, y) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecR S psi"
(*<*)
apply (rule VDMConseq) apply (erule_tac A="(\<lambda> s t . \<forall> r . (s,r):T \<longrightarrow> (t,r):T)" in VDMWhile) 
 apply clarsimp
  apply clarsimp 
  apply (simp add: RELSecR_def) apply clarsimp
done
(*>*)

lemma REL_WhileRight_as_derived:
   "\<lbrakk>{} \<rhd> D : B; {} \<rhd> C : A; 
    \<forall>s t s'. A s t \<longrightarrow> evalB b s \<longrightarrow> ((s', s) \<in> R \<longrightarrow> (s',t) \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (t, s') \<in> T); 
     R = {(s, s'). \<forall>t. B s t \<longrightarrow> (s', t) \<in> T}; S = {(s, s'). (s', s) \<in> T \<and> \<not> evalB b s'};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> RHL R D (While b C) S psi"
(*<*)
apply (unfold RHL_def)
apply rule 
  apply (erule_tac A=A and B=B and C=C and b=b and T=T and S=S in REL_WhileRightL) apply assumption+ 
  apply (erule_tac A=A and B=B and D=D and b=b and T=T and R=R in REL_WhileRightR) apply assumption+ 
done
(*>*)

(*The as-derived rule employs T in its inverse form - the following
rule rectified this, by instantiating T with Inverse T*)

lemma REL_WhileRight:
   "\<lbrakk>{} \<rhd> D : B; {} \<rhd> C : A; 
    \<forall>s' t' s. A s' t' \<longrightarrow> evalB b s' \<longrightarrow> ((s, s') \<in> R \<longrightarrow> (s,t') \<in> R) \<and> ((s, s') \<in> T \<longrightarrow> (s, t') \<in> T); 
     R = {(s, s'). \<forall>t. B s t \<longrightarrow> (t, s') \<in> T}; S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s'};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> RHL R D (While b C) S psi"
(*<*)
apply (unfold RHL_def)
apply rule 
  apply (erule_tac A=A and B=B and C=C and b=b and T="Inverse T" and S=S in REL_WhileRightL) apply assumption+
    apply (simp add: Inverse_def) apply (simp add: Inverse_def) apply (simp add: Inverse_def) apply assumption
  apply (erule_tac A=A and B=B and D=D and b=b and T="Inverse T" and R=R in REL_WhileRightR) apply assumption+ 
  apply (simp add: Inverse_def) apply (simp add: Inverse_def) apply (simp add: Inverse_def)  apply assumption
done
(*>*)

lemma WhileElimLeft_L:
  "\<lbrakk>{} \<rhd> D : A; \<forall> s' t' s . A s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t'):S;
    R = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s};
        phi = (\<lambda> (s,t) . (s,t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R phi"
(*<*)
apply (rule VDMConseq) apply (rule_tac A="\<lambda> s t .(\<not> evalB b s) \<longrightarrow> s = t" in VDMWhile) apply (rule VDMTrue)
     apply simp apply simp 
  apply (simp add:RELSecL_def) 
done
(*>*)

lemma WhileElimLeft_R:
  " \<lbrakk>{} \<rhd> D : A; \<forall> s' t' s . A s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow>(s, t'):S;
      R = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s};
        phi = (\<lambda> (s,t) . (s,t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecR S phi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecR_def) 
done
(*>*)

lemma REL_WhileElimLeft:
 "\<lbrakk>{} \<rhd> D : A; \<forall> s' t' s . A s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> (s, t'):S;
    R = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s};
        phi = (\<lambda> (s,t) . (s,t) \<in> R)\<rbrakk> 
    \<Longrightarrow> RHL R (While b C) D S phi"
(*<*)
apply (unfold RHL_def) 
apply rule
  apply (erule_tac T=T and S=S in WhileElimLeft_L) apply assumption+ 
  apply (erule_tac T=T and R=R and b=b in WhileElimLeft_R) apply assumption+
done
(*>*)

lemma RHLDeadWhileLeft_derivable2:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s}; phi = (\<lambda> (s,t) . (s,t): R )\<rbrakk> 
       \<Longrightarrow> RHL R (While b C) Skip R phi"
(*<*)
apply (rule_tac T=S in REL_WhileElimLeft) apply (rule VDMSkip) 
apply clarsimp apply assumption+ 
done
(*>*)

lemma WhileElimRight_L:
   "\<lbrakk>{} \<rhd> D : B; 
     R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> T)}; 
     S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s'};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecL R psi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecL_def) apply clarsimp
  apply (rule_tac x=s in exI, simp)
done
(*>*)

lemma WhileElimRight_R:
   "\<lbrakk>{} \<rhd> D : B; R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> T)}; S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s'};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecR S psi"
(*<*)
apply (rule VDMConseq)
apply (rule_tac A="(\<lambda> s t . \<forall> r . \<not> evalB b s \<longrightarrow> (r,s):T \<longrightarrow> (r,t):T)" in VDMWhile)  apply (rule VDMTrue)
 apply clarsimp
  apply clarsimp 
  apply (simp add: RELSecR_def) apply clarsimp 
done
(*>*)

lemma REL_WhileElimRight:
   "\<lbrakk>{} \<rhd> D : B; 
     R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> T)}; S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s'};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> RHL R D (While b C) S psi"
(*<*)
apply (unfold RHL_def)
apply rule 
  apply (erule WhileElimRight_L, assumption+)
  apply (erule WhileElimRight_R, assumption+)
done
(*>*)

lemma WhileElimRight_Alternative_L:
  "\<lbrakk>{} \<rhd> D : B; R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> S)}; psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> D : RELSecL R psi"
(*<*)
apply (erule VDMConseq) apply (simp add: RELSecL_def) apply clarsimp
  apply (rule_tac x=s in exI, simp)
done
(*>*)

lemma WhileElimRight_Alternative_R:
  "\<lbrakk>{} \<rhd> D : B; R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> S)}; psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecR S psi"
(*<*)
apply (rule VDMConseq)
apply (rule_tac A="(\<lambda> s t . \<forall> r . \<not> evalB b s \<longrightarrow> (r,s):S \<longrightarrow> (r,t):S)" in VDMWhile)  apply (rule VDMTrue)
 apply clarsimp
  apply clarsimp 
  apply (simp add: RELSecR_def) apply clarsimp 
done
(*>*)

lemma REL_WhileElimRight_Alternative:
   "\<lbrakk>{} \<rhd> D : B; 
     R = {(s, s'). \<not> evalB b s' \<and> (\<forall>t. B s t \<longrightarrow> (t, s') \<in> S)};
     psi = (\<lambda>(s, t). \<exists>r. B r s \<and> (r, t) \<in> R)\<rbrakk>
    \<Longrightarrow> RHL R D (While b C) S psi"
(*<*)
apply (unfold RHL_def)
apply rule 
  apply (erule WhileElimRight_Alternative_L,assumption+)
  apply (erule WhileElimRight_Alternative_R,assumption+)
done
(*>*)

lemma RHLDeadWhileRight_derivable2:
  "\<lbrakk> R = {(s,s') . (s,s'):S \<and> \<not> evalB b s'}; phi = (\<lambda> (s,t) . (s,t): R) \<rbrakk> 
       \<Longrightarrow> RHL R Skip (While b C) R phi"
(*<*)
apply (rule_tac SS=R in RHLSub)
apply (rule REL_WhileElimRight_Alternative) apply (rule VDMSkip) 
apply clarsimp apply simp apply clarsimp apply rule apply rule apply fast apply fast apply clarsimp apply simp
done
(*>*)

lemma WHILE_IterateL:
  "\<lbrakk> {} \<rhd> C : A; {} \<rhd> C' : A'; {} \<rhd> C : RELSecL R phi;

     R = {(s, s'). (s, s') \<in> T \<and> evalB b s \<and> evalB b' s'}; 
     \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b s \<longrightarrow> (t, s') \<in> T;
     
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s, t') \<in> T;
     S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'};
     psi = (\<lambda>(x, y). (x, y) \<in> T \<and> \<not> evalB b x)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R psi"
(*<*)
apply (rule VDMConseq) 
  apply (erule_tac A="\<lambda> x y . \<forall> r . (x,r): T \<longrightarrow> ((y,r):T \<and> \<not> evalB b y)" in VDMWhile) 
        apply simp apply fast 
 apply (simp add: RELSecL_def) 
done
(*>*)

lemma WHILE_IterateR:
  "\<lbrakk> {} \<rhd> C : A; {} \<rhd> C' : A'; {} \<rhd> C' : RELSecR T phi;
    R = {(s, s'). (s, s') \<in> T \<and> evalB b s \<and> evalB b' s'}; 
     \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b s \<longrightarrow> (t, s') \<in> T;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s, t') \<in> T;
     S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'}; 
     psi = (\<lambda>(x, y). (x, y) \<in> T \<and> \<not> evalB b x)\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b' C' : RELSecR S psi"
(*<*)
apply (rule VDMConseq) 
  apply (erule_tac A="\<lambda> x y . \<forall> r . (\<not> evalB b r) \<longrightarrow> (r,x): T \<longrightarrow> ((r,y):T \<and> \<not> evalB b' y)" in VDMWhile) 
      apply simp apply fast 
  apply (simp add: RELSecR_def)  
done
(*>*)

lemma WHILE_Iterate:
  "\<lbrakk> {} \<rhd> C : A; {} \<rhd> C' : A'; RHL R C C' T phi;

     R = {(s, s'). (s, s') \<in> T \<and> evalB b s \<and> evalB b' s'}; 
     \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b s \<longrightarrow> (t, s') \<in> T; 
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> T \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s, t') \<in> T;
     S = {(s, s'). (s, s') \<in> T \<and> \<not> evalB b s \<and> \<not> evalB b' s'};
     psi = (\<lambda> (x, y) . (x,y):T \<and> \<not> evalB b x)\<rbrakk>
    \<Longrightarrow> RHL R (While b C) (While b' C') S psi"
(*<*)
apply (unfold RHL_def, erule conjE)
apply rule
  apply (rule_tac T=T and b=b and b'=b' and A=A and A'=A' in WHILE_IterateL) 
       apply assumption+ 
  apply (rule_tac T=T and b=b and b'=b' and A=A and A'=A' in WHILE_IterateR) 
       apply assumption+ 
done
(*>*) 

lemma While_NonLockstepL:
  "\<lbrakk> {} \<rhd> C' : A'; {} \<rhd> C : A; {} \<rhd> C : RELSecL R1 phi; 
     R1 = {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> evalB b' s'};  
     \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> R \<longrightarrow> evalB b s \<longrightarrow> (t, s') \<in> R;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s, t') \<in> R;
     S = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'};
     psi = (\<lambda>(s, t). (s, t) \<in> R \<and> \<not> evalB b s \<or>
                     (\<exists>r. (\<forall>z. (r, z) \<in> R \<longrightarrow> (s, z) \<in> R) \<and>
                          \<not> evalB b s \<and>
                          \<not> evalB b' t \<and> (\<forall>ta. (\<forall>z. (r, z) \<in> R \<longrightarrow> (ta, z) \<in> R) \<and> \<not> evalB b ta \<longrightarrow> (ta, t) \<in> R \<and> \<not> evalB b' t)))\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b C : RELSecL R psi"
(*<*)
apply (rule REL_UnionL)
prefer 3 apply (subgoal_tac "R = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s} \<union> {(s, s'). (s, s') \<in> R \<and> evalB b s}")
           apply assumption apply (erule thin_rl)+ apply rule apply fastsimp apply fastsimp
    apply (rule_tac S=R in REL_DeadWhileLeftL) apply simp apply simp
apply (rule REL_UnionL)
prefer 3 apply (subgoal_tac "{(s, s'). (s, s') \<in> R \<and> evalB b s} = 
                             {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> \<not> evalB b' s'} \<union> 
                             {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> evalB b' s'}")
           apply assumption apply (erule thin_rl)+ apply rule apply fastsimp apply fastsimp
      apply (rule VDMConseq)
      apply (rule_tac B=A and A="\<lambda> x y . \<forall> z .(x,z):R \<longrightarrow> (y,z):R" in VDMWhile) apply assumption
           apply clarsimp apply clarsimp 
  apply (subgoal_tac "\<forall>s t. (\<forall>z. (s, z) \<in> R \<longrightarrow> (t, z) \<in> R) \<and> \<not> evalB b t \<longrightarrow> 
        RELSecL {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> \<not> evalB b' s'} 
            (\<lambda> (s, t) . (\<exists>r. (\<forall>z. (r, z) \<in> R \<longrightarrow> (s, z) \<in> R) \<and>
                          \<not> evalB b s \<and>
                          \<not> evalB b' t \<and> 
                          (\<forall>ta. (\<forall>z. (r, z) \<in> R \<longrightarrow> (ta, z) \<in> R) \<and> \<not> evalB b ta \<longrightarrow> (ta, t) \<in> R \<and> \<not> evalB b' t))) s t")
    apply assumption apply clarsimp apply (simp add: RELSecL_def) apply clarsimp 
          apply (rule_tac x=s in exI, simp)
  apply (rule VDMConseq)
      apply (rule_tac B=A and A="\<lambda> x y . \<forall> z .(x,z):R \<longrightarrow> (y,z):R" in VDMWhile) apply assumption
           apply clarsimp apply clarsimp
   apply (subgoal_tac "\<forall>s t. (\<forall>z. (s, z) \<in> R \<longrightarrow> (t, z) \<in> R) \<and> \<not> evalB b t \<longrightarrow> RELSecL {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> evalB b' s'} (\<lambda> (s,t) . (s, t) \<in> R \<and> \<not> evalB b s) s t") apply assumption 
    apply (simp add: RELSecL_def) 
apply simp
apply clarsimp apply rule apply fastsimp
done
(*>*)

lemma While_NonLockstepR:
  "\<lbrakk> {} \<rhd> C : A; {} \<rhd> C' : A'; {} \<rhd> C' : RELSecR R phi;
     R1 = {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> evalB b' s'}; 
     \<forall>s t s'. A s t \<longrightarrow> (s, s') \<in> R \<longrightarrow> evalB b s \<longrightarrow> (t, s') \<in> R;
     \<forall>s' t' s. A' s' t' \<longrightarrow> (s, s') \<in> R \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s, t') \<in> R;
     S = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'};
     psi = (\<lambda>(s, t). (s, t) \<in> R \<and> \<not> evalB b s \<or>
                     (\<exists>r. (\<forall>z. (r, z) \<in> R \<longrightarrow> (s, z) \<in> R) \<and>
                          \<not> evalB b s \<and>
                          \<not> evalB b' t \<and> (\<forall>ta. (\<forall>z. (r, z) \<in> R \<longrightarrow> (ta, z) \<in> R) \<and> \<not> evalB b ta \<longrightarrow> (ta, t) \<in> R \<and> \<not> evalB b' t)))\<rbrakk>
    \<Longrightarrow> {} \<rhd> While b' C' : RELSecR S psi"
(*<*)
apply (rule REL_UnionR)
  apply (subgoal_tac "{} \<rhd> While b' C' : RELSecR S (\<lambda>(s, t). (s, t) \<in> R \<and> \<not> evalB b s)", assumption)
    apply (rule VDMConseq)
       apply (rule_tac A="\<lambda> x y . \<forall> z . (\<not> evalB b z) \<longrightarrow> (z,x):R \<longrightarrow> (z,y):R" in VDMWhile) apply assumption apply simp 
         apply clarsimp 
    apply (simp add: RELSecR_def)
  apply (subgoal_tac " {} \<rhd> While b' C' : RELSecR S (\<lambda> (s,t) . (\<exists>r. (\<forall>z. (r, z) \<in> R \<longrightarrow> (s, z) \<in> R) \<and>
                          \<not> evalB b s \<and>
                          \<not> evalB b' t \<and> (\<forall>ta. (\<forall>z. (r, z) \<in> R \<longrightarrow> (ta, z) \<in> R) \<and> \<not> evalB b ta \<longrightarrow> (ta, t) \<in> R \<and> \<not> evalB b' t)) \<or>
                     (s, t) \<in> R \<and> \<not> evalB b s)", assumption)
    apply (rule VDMConseq)
       apply (rule_tac A="\<lambda> x y . \<forall> z .(\<not> evalB b z) \<longrightarrow>  (z,x):R \<longrightarrow> (z,y):R" in VDMWhile) apply assumption apply simp 
         apply clarsimp 
    apply (simp add: RELSecR_def) apply clarsimp
apply simp apply rule apply fastsimp
done
(*>*)

lemma RHL_WHILE_NonLockstep:  
  "\<lbrakk> RHL R1 C C' R phi; {} \<rhd> C: A; {} \<rhd> C': A'; 
     R1 = {(s,s') . (s,s'):R \<and> evalB b s \<and> evalB b' s'};
     \<forall> s t s' . A s t \<longrightarrow> (s,s'):R \<longrightarrow> evalB b s \<longrightarrow> (t,s'):R;
     \<forall> s' t' s . A' s' t' \<longrightarrow> (s,s'):R \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s,t'):R;
     S = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')};
     psi = (\<lambda>(s, t). (s, t) \<in> R \<and> \<not> evalB b s \<or>
                     (\<exists>r. (\<forall>z. (r, z) \<in> R \<longrightarrow> (s, z) \<in> R) \<and>
                          \<not> evalB b s \<and>
                          \<not> evalB b' t \<and> (\<forall>ta. ((\<forall>z. (r, z) \<in> R \<longrightarrow> (ta, z) \<in> R) \<and> \<not> evalB b ta) \<longrightarrow> ((ta, t) \<in> R \<and> \<not> evalB b' t))))\<rbrakk>
  \<Longrightarrow> RHL R (While b C) (While b' C') S psi"
(*<*)
apply (unfold RHL_def, erule conjE)
apply rule
  apply (erule_tac R=R and A=A and A'=A' and b=b and b'=b' in While_NonLockstepL) apply assumption+
  apply (erule_tac R=R and A=A and A'=A' and b=b and b'=b' in While_NonLockstepR) apply assumption+
done
(*>*)

lemma RHL_WHILE_NonLockstep_explicitproof:  
  "\<lbrakk> RHL R1 C C' R phi; R1 = {(s,s') . (s,s'):R \<and> evalB b s \<and> evalB b' s'};
     {} \<rhd> C: A; \<forall> s t s' . A s t \<longrightarrow> (s,s'):R \<longrightarrow> evalB b s \<longrightarrow> (t,s'):R;
     {} \<rhd> C': A'; \<forall> s' t' s . A' s' t' \<longrightarrow> (s,s'):R \<longrightarrow> evalB b' s' \<longrightarrow> (\<not> evalB b s) \<longrightarrow> (s,t'):R;
     S = {(s,s') . (s,s'):R \<and> (\<not> evalB b s) \<and> (\<not> evalB b' s')};
     psi = (\<lambda>(s, t). ((((s, t) \<in> R) \<and> (\<not> (evalB b s))) \<or>
                          ((\<exists>r. ((\<forall>z. (((r, z) \<in> R) \<longrightarrow> ((s, z) \<in> R))) \<and>
                                 ((\<not> (evalB b s)) \<and>
                                  ((\<not> (evalB b' t)) \<and>
                                   (\<forall>ta. (((\<forall>z. (((r, z) \<in> R) \<longrightarrow> ((ta, z) \<in> R))) \<and> (\<not> (evalB b ta))) \<longrightarrow>
                                          (((ta, t) \<in> R) \<and> (\<not> (evalB b' t))))))))) \<or>
                           (((s, t) \<in> R) \<and> (\<not> (evalB b s))))))\<rbrakk> \<Longrightarrow>
   RHL R (While b C) (While b' C') S psi"
(*<*)
apply (rule RHLUnion)
prefer 3 apply (subgoal_tac "R = {(s, s'). (s, s') \<in> R \<and> \<not> evalB b s} \<union> {(s, s'). (s, s') \<in> R \<and> evalB b s}")
           apply assumption apply (erule thin_rl)+ apply rule apply fastsimp apply fastsimp
    apply (rule_tac T=R in REL_WhileElimLeft)
      apply (erule_tac A="\<lambda> x y . \<forall> z . (\<not> evalB b z) \<longrightarrow> (z,x):R \<longrightarrow> (z,y):R" in VDMWhile) apply clarsimp apply clarsimp
    apply clarsimp 
    apply simp
    apply simp
apply (rule RHLUnion)
prefer 3 apply (subgoal_tac "{(s, s'). (s, s') \<in> R \<and> evalB b s} = 
                             {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> \<not> evalB b' s'} \<union> 
                             {(s, s'). (s, s') \<in> R \<and> evalB b s \<and> evalB b' s'}")
           apply assumption apply (erule thin_rl)+ apply rule apply fastsimp apply fastsimp
    apply (rule_tac SS="{(s, s'). (s, s') \<in> R \<and> \<not> evalB b s \<and> \<not> evalB b' s'}" in RHLSub)
    apply (rule REL_WhileElimRight_Alternative)
      apply (erule_tac A="\<lambda> x y . \<forall> z . (x,z):R \<longrightarrow> (y,z):R" in VDMWhile) apply clarsimp apply clarsimp 
      apply simp apply simp
    apply clarsimp 
    apply simp
apply (rule_tac T=R and A=A and A'=A' and b=b and b'=b' and C=C and C'=C' and phi=phi in WHILE_Iterate)
  apply assumption+
  apply simp
  apply simp
  apply assumption+
  apply simp
apply simp
apply simp 
done
(*>*)

lemma RHLCBInvTL_valid:
 "\<lbrakk>RHL RR (Iff b C C') D S phi; R = {(s,s') . (s,s') : RR \<and> evalB b s}\<rbrakk>
 \<Longrightarrow> RHL R C D S phi"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply (drule VDM_Sound)
apply (rule VDM_Complete)
apply (simp add: v_valid_def vdmctxt_valid_def)
apply (simp add: vdm_valid_def RELSecL_def Sem_def) apply clarsimp
apply (erule_tac x=s in allE, erule_tac x=t in allE, erule impE)
  apply (rule, erule SemTrue) apply assumption
apply fast
done
(*>*)

lemma RHLCBInvFL_valid:
 "\<lbrakk>RHL RR (Iff b C C') D S phi; R = {(s,s') . (s,s') : RR \<and> \<not> evalB b s}\<rbrakk>
 \<Longrightarrow> RHL R C' D S phi"
(*<*)
apply (simp add: RHL_def) apply clarsimp
apply (drule VDM_Sound)
apply (rule VDM_Complete)
apply (simp add: v_valid_def vdmctxt_valid_def)
apply (simp add: vdm_valid_def RELSecL_def Sem_def) apply clarsimp
apply (erule_tac x=s in allE, erule_tac x=t in allE, erule impE)
  apply (rule, erule SemFalse) apply assumption
apply fast
done
(*>*)

(*<*)
end
(*>*)