theory VDMSound = VDM:

subsection{*Soundness of VDM logic*}
(*<*)
lemma MAX:"Suc (max k m) \<le> n \<Longrightarrow> k \<le> n \<and> m \<le> n"
by (simp add: max_def, case_tac "k \<le> m", simp+)
(*>*)

text{*We define a relatived (height-indexed) variant of
validity\ldots*}

constdefs vdm_validn :: "nat \<Rightarrow> IMP \<Rightarrow> VDMAssn \<Rightarrow> bool"
                        (" \<Turnstile>\<^sub>_ _ : _ " [100,100,100] 100)
"vdm_validn n C A \<equiv> 
   (\<forall> m . m \<le> n \<longrightarrow> (\<forall> s t . (s,C \<longrightarrow>\<^sub>m t) \<longrightarrow> A s t))"

text{*\ldots and prove some obvious relationships.*}

lemma VDM_valid_validn: "\<Turnstile> C:A \<Longrightarrow> \<Turnstile>\<^sub>n C:A"
(*<*)
by (simp add: vdm_valid_def vdm_validn_def Sem_def, fastsimp)
(*>*)

lemma VDM_validn_valid: "(\<forall> n . \<Turnstile>\<^sub>n C:A) \<Longrightarrow> \<Turnstile> C:A"
(*<*)
by (simp add: vdm_valid_def vdm_validn_def Sem_def, fastsimp)
(*>*)

lemma VDM_lowerm: "\<lbrakk> \<Turnstile>\<^sub>n C:A; m \<le> n \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>m C:A"
(*<*)
apply (simp add: vdm_validn_def, clarsimp)
apply (erule_tac x="ma" in allE, simp)
done
(*>*)

text{*Likewise for contexts\ldots*}

constdefs vdmctxt_validn :: "nat \<Rightarrow> VDMCtxt \<Rightarrow> bool"
                         (" \<Turnstile>\<^sub>_ _  " [100,100] 100)
"vdmctxt_validn n G \<equiv>
   \<forall> m . m \<le> n \<longrightarrow> (\<forall> A. A \<in> G \<longrightarrow> ( \<Turnstile>\<^sub>m Call : A) )"

lemma Ctxt_valid_validn: "\<Turnstile> G \<Longrightarrow> \<Turnstile>\<^sub>n G"
(*<*)
apply (simp add: vdmctxt_valid_def vdmctxt_validn_def, clarsimp)
apply (rule VDM_valid_validn) apply fast
done
(*>*)

lemma Ctxt_validn_valid: "(\<forall> n . \<Turnstile>\<^sub>n G) \<Longrightarrow> \<Turnstile> G"
(*<*)
apply (simp add: vdmctxt_valid_def vdmctxt_validn_def, clarsimp)
apply (rule VDM_validn_valid) apply fast
done
(*>*)

lemma Ctxt_lowerm: "\<lbrakk> \<Turnstile>\<^sub>n G; m < n \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>m G"
(*<*)
by (simp add: vdmctxt_validn_def)
(*>*)

text{*\ldots and finally for judgements.*}

constdefs v_validn::"(VDMAssn set) \<Rightarrow> nat \<Rightarrow> IMP \<Rightarrow>
                      VDMAssn \<Rightarrow> bool"
                    ("_ \<Turnstile>\<^sub>_ _ : _" [100,100,100,100] 100)
"v_validn G n C A \<equiv> \<Turnstile>\<^sub>n G \<longrightarrow> \<Turnstile>\<^sub>n C:A"

lemma validn_valid: "(\<forall> n . G \<Turnstile>\<^sub>n C : A) \<Longrightarrow> G \<Turnstile> C : A"
(*<*)
apply (simp add: v_valid_def v_validn_def, clarsimp)
apply (rule VDM_validn_valid, clarsimp) 
apply (erule_tac x=n in allE, erule mp)
apply (erule Ctxt_valid_validn)
done
(*>*)

lemma ctxt_consn: "\<lbrakk> \<Turnstile>\<^sub>n G;  \<Turnstile>\<^sub>n Call:A \<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>n ({A} \<union> G)"
(*<*)
apply (simp add: vdmctxt_validn_def)  apply clarsimp 
apply (erule_tac x=m in allE, clarsimp) 
apply (erule VDM_lowerm) apply assumption
done
(*>*)

(*<*)
constdefs SoundWhileProp::"nat \<Rightarrow> (VDMAssn set) \<Rightarrow> IMP \<Rightarrow> VDMAssn \<Rightarrow> BExpr \<Rightarrow> VDMAssn \<Rightarrow> bool"
"SoundWhileProp n G C B b A \<equiv>
   (\<forall>m. G \<Turnstile>\<^sub>m C : B) \<longrightarrow> (\<forall>s. (\<not> evalB b s) \<longrightarrow> A s s) \<longrightarrow>
    (\<forall>s. evalB b s \<longrightarrow> (\<forall>r. B s r \<longrightarrow> (\<forall>t. A r t \<longrightarrow> A s t))) \<longrightarrow>
   G \<Turnstile>\<^sub>n (While b C) : (\<lambda>s t. A s t \<and> \<not> evalB b t)"

lemma SoundWhileAux[rule_format]: "SoundWhileProp n G C B b A"
apply (induct n)
apply (simp_all add: SoundWhileProp_def)
apply clarsimp apply (simp add: v_validn_def vdm_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp apply (simp add: v_validn_def vdm_validn_def, clarsimp)  
  apply (erule Sem_eval_cases)
  prefer 2 apply clarsimp 
  apply clarsimp 
   apply (erule_tac x=n in allE, erule impE) apply (erule Ctxt_lowerm) apply simp
   apply (rotate_tac -1)
   apply (erule_tac x=ma in allE, clarsimp) 
   apply(erule impE) apply (erule Ctxt_lowerm) apply simp 
   apply (erule_tac x=na in allE, clarsimp) apply fast
done
(*>*)

text{*An auxiliary lemma stating the soundness of the while rule. Its
proof is by induction on $n$.*}

lemma VDMWhileSound[rule_format]:
  "(\<forall>m. G \<Turnstile>\<^sub>m C : B) \<longrightarrow> (\<forall>s. (\<not> evalB b s) \<longrightarrow> A s s) \<longrightarrow>
    (\<forall>s. evalB b s \<longrightarrow> (\<forall>r. B s r \<longrightarrow> (\<forall>t. A r t \<longrightarrow> A s t))) \<longrightarrow>
   G \<Turnstile>\<^sub>n (While b C) : (\<lambda>s t. A s t \<and> \<not> evalB b t)"
(*<*)
apply (subgoal_tac "SoundWhileProp n G C B b A")
  apply (simp add: SoundWhileProp_def)
apply (rule SoundWhileAux)
done 
(*>*)

text{*The above is the auxiliary lemma for the proof rule
VDMWhile. For rule VDMWhile\_Hoare, the lemma is as follows - the
proof is essentially identical.*}

lemma VDMWhile_HoareSound[rule_format]:
   "(\<forall>n. G \<Turnstile>\<^sub>n C : (\<lambda>u ua. evalB b u \<longrightarrow> (\<forall>r. I u r \<longrightarrow> I ua r))) \<longrightarrow>
       (G \<Turnstile>\<^sub>n (While b C) : (\<lambda>u ua. \<forall>r. I u r \<longrightarrow> I ua r \<and> \<not> evalB b ua))"
(*<*)
apply (induct n)
apply clarsimp apply (simp add: v_validn_def vdm_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp apply (simp add: v_validn_def vdm_validn_def, clarsimp)  
  apply (erule Sem_eval_cases)
  prefer 2 apply clarsimp 
  apply clarsimp 
   apply (erule_tac x=n in allE, erule impE) apply (erule Ctxt_lowerm) apply simp
   apply (rotate_tac -1)
   apply (erule_tac x=ma in allE, clarsimp) 
   apply(erule impE) apply (erule Ctxt_lowerm) apply simp 
   apply (erule_tac x=na in allE, clarsimp) apply fast
done
(*>*)

text{*Similarly, an auxiliary lemma for procedure invocations.*}

lemma SoundCall[rule_format]:
"\<lbrakk>\<forall>n. \<Turnstile>\<^sub>n ({A} \<union> G) \<longrightarrow> \<Turnstile>\<^sub>n body : A\<rbrakk> \<Longrightarrow> \<Turnstile>\<^sub>n G \<longrightarrow> \<Turnstile>\<^sub>n Call : A"
(*<*)
apply (induct_tac n)
apply (simp add: vdm_validn_def, clarsimp) 
  apply (drule Sem_no_zero_height_derivs) apply simp 
apply clarsimp
  apply (drule Ctxt_lowerm) apply (subgoal_tac "n < Suc n", assumption) apply simp apply clarsimp
  apply (drule ctxt_consn) apply assumption
  apply (simp add: vdm_validn_def, clarsimp)
  apply (erule Sem_eval_cases, clarsimp) 
done
(*>*)

text{*The heart of the soundness proof is the following lemma which is
proven by induction on the judgement $G \rhd C :A$. The cases for
while and call invoke the above lemmas.*}

lemma VDM_Sound_n: "G \<rhd> C: A \<Longrightarrow> (\<forall> n . G \<Turnstile>\<^sub>n C:A)"
(*<*)
apply (erule VDM_proof.induct)
apply (simp add: v_validn_def vdm_validn_def)
  apply(clarsimp, erule Sem_eval_cases, simp)
apply (simp add: v_validn_def vdm_validn_def)
  apply(clarsimp, erule Sem_eval_cases, simp)
apply (simp add: v_validn_def vdm_validn_def)
  apply(clarsimp, erule Sem_eval_cases, clarsimp)
  apply (drule MAX, clarsimp)
  apply (erule_tac x=n in allE, clarsimp, rotate_tac -1, erule_tac x=na in allE, clarsimp)
  apply (erule_tac x=n in allE, clarsimp, rotate_tac -1, erule_tac x=ma in allE, clarsimp)
  apply (rule_tac x=r in exI, fast)
apply (simp add: v_validn_def vdm_validn_def)
  apply(clarsimp, erule Sem_eval_cases, clarsimp)
   apply (erule thin_rl, rotate_tac 1, erule thin_rl, erule thin_rl)
     apply (erule_tac x=n in allE, clarsimp, erule_tac x=na in allE, clarsimp)
   apply (erule thin_rl, erule thin_rl)
     apply (erule_tac x=n in allE, clarsimp, erule_tac x=na in allE, clarsimp)
apply clarsimp
  apply (rule VDMWhile_HoareSound) apply fast 
apply (simp add: v_validn_def, clarsimp)
  apply (rule_tac G=G in SoundCall) apply fastsimp apply assumption 
apply (simp add: vdmctxt_validn_def v_validn_def) 
apply (simp add: v_validn_def vdm_validn_def) 
apply (simp add: v_validn_def vdm_validn_def) 
apply (simp add: v_validn_def vdm_validn_def) 
done
(*>*)

text{*Thus, we obtain soundness in contexts,\ldots*}

theorem VDM_Sound: "G \<rhd> C: A \<Longrightarrow> G \<Turnstile> C:A"
(*<*)
by (drule VDM_Sound_n, erule validn_valid) 
(*>*)

text{*\ldots a simple corollary of which is soundness w.r.t.~an empty
context.*}

lemma VDM_Sound_emptyCtxt:"{} \<rhd> C : A \<Longrightarrow> \<Turnstile> C : A"
(*<*)
apply (drule VDM_Sound)
apply (simp add: v_valid_def, erule mp) 
apply (simp add: vdmctxt_valid_def)
done
(*>*)


text{*Combining this result with the rule {\sc MUTREC} yields that
finite verified contexts are valid.*}

lemma Ctxt_verified_valid: "\<lbrakk>verified G; finite G\<rbrakk> \<Longrightarrow> vdmctxt_valid G"
(*<*)
apply (subgoal_tac "(\<forall> A . A:G \<longrightarrow> G \<rhd> body : A)")
prefer 2 apply (simp add: verified_def)
apply (simp add: vdmctxt_valid_def, clarsimp)
apply (erule_tac x=A in allE, simp)
apply (rule VDM_Sound_emptyCtxt)
apply (subgoal_tac "card G = card G")
apply (rule MUTREC, assumption)  apply assumption+ apply simp
(*  apply (simp add: card_def)*)
done
(*>*)

(*<*)
end
(*>*)