(*<*)
theory VDMinHoare = HoareMultiCall + VDM:
(*>*)

subsection{*Interpretation of VDM logic in Hoare logic*}

(*<*) text{*The following stuff concerns semantic validity. Since
we're aiming at formal derivability, it's not really needed, so we
don't include it in the document.*}

lemma "\<lbrakk> \<forall>r. \<Turnstile> { (\<lambda>s. s = r) } C1 { A1 r }; \<forall>r. \<Turnstile> { (\<lambda>s. s = r) } C2 { A2 r }\<rbrakk>
       \<Longrightarrow> \<Turnstile> { (\<lambda>s. s = r) } Comp C1 C2 { (\<lambda>t. \<exists>ra. A1 r ra \<and> A2 ra t) }"
apply (simp add:Hoare_valid_def Sem_def, clarsimp)
apply (erule Semn.elims, simp_all) apply clarsimp
apply (erule_tac x=r in allE, erule_tac x=ra in allE, erule impE, rule, assumption)
apply (erule_tac x=ra in allE, erule_tac x=ta in allE, erule impE, rule, assumption)
apply (rule_tac x=ra in exI, simp)
done
        
constdefs While_Prop::"IMP \<Rightarrow> BExpr \<Rightarrow> VDMAssn \<Rightarrow> bool"
"While_Prop C b I \<equiv> \<forall>r t. (\<exists>n.  r , C \<longrightarrow>\<^sub>n  t ) \<longrightarrow> evalB b r \<longrightarrow> (\<forall>ra. I r ra \<longrightarrow> I t ra)"

lemma While_Aux[rule_format]: 
"\<forall> n s r t . s , While b C \<longrightarrow>\<^sub>n  t \<longrightarrow> I s r \<longrightarrow> n <= m \<longrightarrow> While_Prop C b I \<longrightarrow>  
        (I t r \<and> \<not> evalB b t)"
apply (induct m)
apply clarsimp
  apply (drule Sem_no_zero_height_derivs) apply simp
apply clarsimp
  apply (erule Sem_eval_cases)
  apply clarsimp 
  prefer 2 apply clarsimp 
  apply (subgoal_tac "I ra r")
   prefer 2 apply (erule thin_rl) apply (simp add: While_Prop_def) 
            apply (erule_tac x=s in allE, rotate_tac -1) 
            apply (erule_tac x=ra in allE, rotate_tac -1) 
            apply (erule impE) apply (rule, assumption) 
            apply clarsimp
  apply (erule_tac x=ma in allE) apply clarsimp
done
(*>*)

text{*We combine pre- and post-conditions of Hoare logics to a notion
of specifications*}

types Spec = "HoareAssn \<times> HoareAssn"

text{*and interprete a VDM assertion as a state-indexed specification
(i.e. a function from states to specifications.*}

constdefs encode::"VDMAssn \<Rightarrow> State \<Rightarrow> Spec"
"encode A r \<equiv> (\<lambda> s . s=r, \<lambda> t . A r t)"

text{*A small variation of this is to interprete VDM assertions
directly as specification sets, as in the paper:*}

constdefs encode1::"VDMAssn \<Rightarrow> Spec set"
"encode1 A \<equiv> { X . (\<exists> r . X = (\<lambda> s . s=r, \<lambda> t . A r t))}"

text{*We have the following relationship*}
lemma "(\<exists> r . encode A r = (P,Q)) = ((P,Q): encode1 A)" 
(*<*)
apply (simp add: encode_def encode1_def)
apply rule apply fast apply fast
done
(*>*)

text{*The interpretation of a VDM context is the union of the
interpretation of its constitent VDM specifications.*}

constdefs CtxtEncode :: "VDMCtxt \<Rightarrow> HoareCtxt"
"CtxtEncode G \<equiv> { x . \<exists> A r . A : G \<and> x = encode A r}"

(*<*) text{*Again some stuff regarding validity (doesn't deal with
procedures yet) - it's not needed so we suppress document
generation.*}

constdefs Call_Prop::"VDMAssn \<Rightarrow> VDMCtxt \<Rightarrow> bool"
"Call_Prop A G \<equiv>  \<Turnstile> CtxtEncode (insert A G)  \<longrightarrow> (\<forall>r t. (\<exists>n.  r , body \<longrightarrow>\<^sub>n  t ) \<longrightarrow> A r t)"

lemma "\<forall> n s t. n <= m \<longrightarrow> s , Call \<longrightarrow>\<^sub>n  t \<longrightarrow> Call_Prop A G \<longrightarrow> \<Turnstile> CtxtEncode G \<longrightarrow> A s t"
apply (induct m)
apply clarsimp 
  apply (drule Sem_no_zero_height_derivs) apply simp
apply clarsimp
  apply (erule Sem_eval_cases)
  apply clarsimp 
  apply (subgoal_tac "\<Turnstile> CtxtEncode (insert A G)")
    apply (simp add: Call_Prop_def)
    apply (erule_tac x=s in allE, erule_tac x=t in allE, erule mp)
    apply (rule, assumption)
  apply (simp add: CtxtEncode_def Ctxt_valid_def, clarsimp)
  apply (erule disjE) apply clarsimp
oops
(*ToDo: Usual trick with relativised context validity*)

lemma 
"G \<rhd> C :A \<Longrightarrow> \<forall> r . (CtxtEncode G) \<Turnstile> {(fst (encode A r))} C {(snd (encode A r))}"
apply (erule VDM_proof.induct)
apply (simp_all add: encode_def valid_def Hoare_valid_def Sem_def)
apply clarsimp apply (erule Sem_eval_cases) apply assumption 
apply clarsimp apply (erule Sem_eval_cases) apply assumption 
apply clarsimp apply (erule Sem_eval_cases) apply fast 
apply clarsimp apply (erule Sem_eval_cases) 
  apply (rule, clarsimp) apply fast apply clarsimp apply fast 
apply clarsimp 
  apply (erule_tac m=n in While_Aux) apply assumption apply simp 
    apply (simp add: While_Prop_def)
defer (*Call*)
defer (*Ax*)
(*apply clarsimp 
  apply (subgoal_tac "\<Turnstile> CtxtEncode (insert A G)", clarsimp)
    apply (erule_tac x=r in allE, erule_tac x=t in allE, erule mp)
    apply (erule Sem_eval_cases) apply fast 
  apply (erule thin_rl) apply (rotate_tac -1, erule thin_rl)
  apply (simp add: CtxtEncode_def Ctxt_valid_def)  apply clarsimp 
  apply (simp add: encode_def) apply clarsimp 
  apply (erule disjE) apply clarsimp 
    apply (erule_tac x="fst (encode A r)" in allE)
  apply (rotate_tac -1, erule_tac x="snd (encode A r)" in allE, erule impE)
  *)
apply clarsimp
  apply (simp add: CtxtEncode_def Hoare_valid_def Ctxt_valid_def Sem_def)
  (*apply (erule_tac x="fst (encode A r)" in allE) 
  apply (erule_tac x="snd (encode A r)" in allE)
  apply (erule impE) apply (rule_tac x=A in exI, simp) apply (rule_tac x=r in exI, simp)
  apply (simp add: encode_def) apply fast*)
oops
(*>*)

text{*The following result shows that the interpretation is sound,
i.e. that the VDM logic can indeed be embedded in the Hoare logic.*}

lemma VDM_derivable[rule_format]:
"G \<rhd> C :A \<Longrightarrow> \<forall> r . (CtxtEncode G) \<rhd> {(fst (encode A r))} C {(snd (encode A r))}"
(*<*)
apply (erule VDM_proof.induct)
apply (simp_all add: encode_def)
(*Skip*)
apply clarsimp apply (rule HoareSkip)  
(*Assign*)
apply clarsimp 
  apply (rule_tac Q="\<lambda>t. t = update r x (evalE e r)" in HoareConseq)
  apply (rule HoareAssign) apply simp apply simp 
(*Comp*)
apply clarsimp
  apply (erule_tac x=r in allE) 
  apply (subgoal_tac "CtxtEncode G \<rhd> {(\<lambda> s . s=r)} C1 {(\<lambda> t . \<exists> y . A1 r y \<and> y=t)}")
  prefer 2 apply (erule HoareConseq) apply simp apply clarsimp 
  apply (rotate_tac -2, erule thin_rl)
  apply (erule HoareComp)
  apply (rule extractexistsState) apply clarsimp
  apply (erule_tac x=x in allE)
  apply (drule_tac P="\<lambda> a . A1 r x \<and> x = a" in HoareCons)
  apply (erule HoareConseq) apply clarsimp apply clarsimp
(*Iff*) 
  apply clarsimp 
  apply (erule_tac x=r in allE)
  apply (erule_tac x=r in allE)
  apply (erule HoareIff2) apply assumption
  apply clarsimp
(*While*)
  apply clarsimp apply (erule thin_rl)
  apply (rule_tac P="\<lambda> u . u=r" and Q="\<lambda> t . (\<forall> ra . I r ra \<longrightarrow> I t ra) \<and> \<not> evalB b t" in HoareConseq)
     prefer 2 apply simp prefer 2 apply simp
  apply (rule_tac Q="(\<lambda>t. (\<forall>ra. I r ra \<longrightarrow> I t ra) \<and> \<not> evalB b t)" in HoareConseq)
  apply (rule Hoare_While_derivable)
    apply (subgoal_tac "(CtxtEncode G) \<rhd> { ( \<lambda> t . ( \<exists> u . (( \<forall> ra. I r ra \<longrightarrow> I t ra) \<and> evalB b t) \<and> u=t)) } C 
                                          { ( \<lambda> s . (\<exists> u . (( \<forall> ra. I r ra \<longrightarrow> I s ra))))}")
    apply (erule HoareConseq) apply clarsimp apply clarsimp
    apply (rule extractexistsState) apply clarsimp 
      apply (erule_tac x=x in allE)
       apply (erule HoareConseqStrong) apply clarsimp  apply clarsimp 
    apply clarsimp
(*Call*)
  apply clarsimp apply (erule thin_rl) 
  apply (rule_tac GG="CtxtEncode {A}" in HoareMultiCall) 
    apply (simp add: CtxtEncode_def encode_def, clarsimp) 
      apply (erule_tac x=r in allE)
      apply (erule HoareCtxtSub) apply rule apply clarsimp 
      apply (erule disjE, clarsimp) apply (rule_tac x=r in exI, simp) 
      apply (erule_tac x=Aa in allE, clarsimp) apply (erule_tac x=r in allE, clarsimp)
    apply (simp add: CtxtEncode_def encode_def) apply (rule_tac x=r in exI, simp)
(*Ax*)
  apply clarsimp apply (rule HoareAx) apply (simp add: CtxtEncode_def encode_def) 
  apply (rule_tac x=A in exI, simp) apply (rule_tac x=r in exI, simp)
(*Conseq*)
  apply clarsimp 
  apply (erule_tac x=r in allE)
  apply (erule HoareConseq) apply simp apply simp
(*VDMTrue*)
  apply clarsimp apply (rule HoareTrue)
(*VDMinter*)
apply clarsimp apply (erule_tac x=r in allE)+ 
  apply (erule HoareInter) apply assumption 
done

text{*In the variant of the encoding used in the paper, this looks as
follows.*}

constdefs CtxtEncode1 :: "VDMCtxt \<Rightarrow> HoareCtxt"
"CtxtEncode1 G \<equiv> { x . \<exists> A . A : G \<and> x : encode1 A}"

lemma VDM_derivable1[rule_format]:
"G \<rhd> C :A \<Longrightarrow> \<forall> P Q . (P,Q): encode1 A \<longrightarrow> (CtxtEncode1 G) \<rhd> {P} C {Q}"
(*<*)
apply (erule VDM_proof.induct)
apply (simp_all add: encode1_def)
(*Skip*)
apply clarsimp apply (rule HoareSkip)  
(*Assign*)
apply clarsimp 
  apply (rule_tac Q="\<lambda>t. t = update r x (evalE e r)" in HoareConseq)
  apply (rule HoareAssign) apply simp apply simp 
(*Comp*)
apply clarsimp
  apply (erule_tac x="\<lambda> s . s=r" in allE, rotate_tac -1)
  apply (erule_tac x="A1 r" in allE, rotate_tac -1)
  apply (erule impE) apply (rule_tac x=r in exI, simp)
  apply (subgoal_tac "CtxtEncode1 G \<rhd> {(\<lambda> s . s=r)} C1 {(\<lambda> t . \<exists> y . A1 r y \<and> y=t)}")
  prefer 2 apply (erule HoareConseq) apply simp apply clarsimp 
  apply (rotate_tac -2, erule thin_rl)
  apply (erule HoareComp)
  apply (rule extractexistsState) apply clarsimp
  apply (erule_tac x="\<lambda> s . s=x" in allE, rotate_tac -1)
  apply (erule_tac x="A2 x" in allE, rotate_tac -1)
  apply (erule impE) apply (rule_tac x=x in exI, simp)
  apply (drule_tac P="\<lambda> a . A1 r x \<and> x = a" in HoareCons)
  apply (erule HoareConseq) apply clarsimp apply clarsimp
(*Iff*) 
  apply clarsimp 
  apply (erule_tac x="\<lambda> s . s=r" in allE, rotate_tac -1)
  apply (erule_tac x="A r" in allE, rotate_tac -1)
  apply (erule impE) apply (rule_tac x=r in exI, simp)
  apply (erule_tac x="\<lambda> s . s=r" in allE, rotate_tac -1)
  apply (erule_tac x="B r" in allE, rotate_tac -1)
  apply (erule impE) apply (rule_tac x=r in exI, simp)
  apply (erule HoareIff2) apply assumption
  apply clarsimp
(*While*)
  apply clarsimp apply (erule thin_rl)
  apply (rule_tac P="\<lambda> u . u=r" and Q="\<lambda> t . (\<forall> ra . I r ra \<longrightarrow> I t ra) \<and> \<not> evalB b t" in HoareConseq)
     prefer 2 apply simp prefer 2 apply simp
  apply (rule_tac Q="(\<lambda>t. (\<forall>ra. I r ra \<longrightarrow> I t ra) \<and> \<not> evalB b t)" in HoareConseq)
  apply (rule Hoare_While_derivable)
    apply (subgoal_tac "(CtxtEncode1 G) \<rhd> { ( \<lambda> t . ( \<exists> u . (( \<forall> ra. I r ra \<longrightarrow> I t ra) \<and> evalB b t) \<and> u=t)) } C 
                                          { ( \<lambda> s . (\<exists> u . (( \<forall> ra. I r ra \<longrightarrow> I s ra))))}")
    apply (erule HoareConseq) apply clarsimp apply clarsimp
    apply (rule extractexistsState) apply clarsimp 
      apply (erule_tac x="\<lambda> s . s=x" in allE, rotate_tac -1)
      apply (erule_tac x="\<lambda>t. evalB b x \<longrightarrow> (\<forall>ra. I x ra \<longrightarrow> I t ra)" in allE, rotate_tac -1)
      apply (erule impE) apply (rule_tac x=x in exI, simp)
      apply (erule HoareConseqStrong) apply clarsimp apply clarsimp 
    apply clarsimp
(*MultiCall*)
  apply clarsimp apply (erule thin_rl) 
  apply (rule_tac GG="CtxtEncode {A}" in HoareMultiCall) 
    apply (simp add: CtxtEncode_def encode_def, clarsimp) 
      apply (erule_tac x="\<lambda> s . s=r" in allE) apply (erule_tac x="A r" in allE)
      apply (erule impE) apply (rule_tac x=r in exI, simp)
      apply (erule HoareCtxtSub) apply rule apply clarsimp
      apply (simp add: CtxtEncode1_def) apply clarsimp
      apply (erule disjE, simp add: encode1_def, clarsimp) 
    apply (simp add: CtxtEncode_def encode_def) apply (rule_tac x=r in exI, simp)
(*Ax*)
  apply clarsimp apply (rule HoareAx) apply (simp add: CtxtEncode1_def) apply (rule_tac x=A in exI, simp)
  apply (simp add: encode1_def) apply (rule_tac x=r in exI, simp)
(*Conseq*)
  apply clarsimp
  apply (erule_tac x="\<lambda> s . s=r" in allE, rotate_tac -1)
  apply (erule_tac x="A r" in allE, rotate_tac -1)
  apply (erule impE) apply (rule_tac x=r in exI, simp)
  apply (erule HoareConseq) apply simp apply simp
(*VDMTrue*)
  apply clarsimp  apply (rule HoareTrue)
(*VDMInter*)
  apply clarsimp 
  apply (erule_tac x="\<lambda>u. u = r" in allE, erule_tac x="A1 r" in allE, erule impE)
    apply (rule_tac x=r in exI, simp)
  apply (erule_tac x="\<lambda>u. u = r" in allE, erule_tac x="A2 r" in allE, erule impE)
    apply (rule_tac x=r in exI, simp)
  apply (erule HoareInter) apply assumption 
done
(*>*)

lemma VDM_Sound_via_Hoare: "G \<rhd> C : A \<Longrightarrow> v_valid G C A"
(*<*) 
apply (subgoal_tac "\<forall> r . CtxtEncode G  \<Turnstile> { fst (encode A r) } C { snd (encode A r) }")
  prefer 2 apply clarsimp apply (rule Hoare_Sound) apply (erule VDM_derivable)
apply (erule thin_rl)
apply (simp add: v_valid_def valid_def, clarsimp)
apply (erule impE)
   apply (simp add: vdmctxt_valid_def Ctxt_valid_def, clarsimp)
   apply (simp add: CtxtEncode_def, clarsimp)
   apply (erule_tac x=A in allE, clarsimp)
   apply (simp add: Hoare_valid_def vdm_valid_def, clarsimp)
   apply (erule_tac x=s in allE, erule_tac x=t in allE, clarsimp)
   apply (simp add: encode_def)
apply (simp add: vdm_valid_def Hoare_valid_def, clarsimp) 
   apply (erule_tac x=s in allE, erule_tac x=s in allE, erule_tac x=t in allE, clarsimp)
   apply (simp add: encode_def)
done
(*>*)

(*<*)
end
(*>*)