(*<*)
theory VS = VDMComplete:
(*Authors: Lennart Beringer and Martin Hofmann, LMU Munich 2007*)
(*>*)

section{*Base-line noninterference*}

text{*\label{sec:BaseLineNI} We introduce logical self-composition for
non-interference and show how to interprete the type system of
Volpano, Smith and Irvine. Apart from the final subsection, this
section is unchanged from the previous formalization and hence
includes the treatment of the procedure.*}

subsection{*Basic definitions*}

text{*We have the usual two security types.*}

datatype TP = low | high

text{*A (global) context assigns a security type to each program
variable.*}

consts CONTEXT :: "Var \<Rightarrow> TP"

text{*Next, we define when two states are considered (low)
equivalent.*}

constdefs twiddle::"State \<Rightarrow> State \<Rightarrow> bool"
"twiddle s ss \<equiv> \<forall> x. CONTEXT x = low \<longrightarrow> s x = ss x"

text{*A command $C$ is \emph{secure} if the low equivalence of any two
initial states entails the equivalence of the corresponding final
states.*}

constdefs secure::"IMP \<Rightarrow> bool"
"secure C \<equiv> \<forall> s t ss tt . twiddle s t \<longrightarrow> (s,C \<Down> ss) \<longrightarrow> 
                          (t,C \<Down> tt) \<longrightarrow> twiddle ss tt"

text {*Here is the definition of the assertion transformer
$\mathit{Sec}$ \ldots*}

types TT = "(State \<times> State) \<Rightarrow> bool"

constdefs Sec :: "TT \<Rightarrow> VDMAssn"
"Sec phi s t \<equiv> (\<forall> r . twiddle s r \<longrightarrow> phi(t, r)) \<and> 
             (\<forall> r . phi(r ,s) \<longrightarrow> twiddle r t)"

text{*\ldots and the proofs of its characteristic property,
Proposition 1.*}

lemma Prop1A : "vdm_valid C (Sec Phi) \<Longrightarrow> secure C"
by (simp add: vdm_valid_def secure_def Sec_def)


lemma Prop1B:
  "secure C \<Longrightarrow> vdm_valid C (Sec (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> twiddle s t))"
(*<*)
apply (simp add: vdm_valid_def Sec_def) 
apply clarsimp
apply (unfold secure_def)
apply rule
apply (rule, rule) apply (rule_tac x=s in exI, rule, assumption+)
apply (rule, rule, erule exE, erule conjE) apply fast
done
(*>*)

lemma Prop1BB:"secure C \<Longrightarrow> \<exists> Phi . vdm_valid C (Sec Phi)"
(*<*)
by (drule Prop1B, fast)
(*>*)

lemma Prop1:"(secure C)
             = (vdm_valid C (Sec (\<lambda> (r, t) . \<exists> s . (s , C \<Down> r) \<and> twiddle s t)))"
(*<*)
apply rule
apply (erule Prop1B)
apply (erule Prop1A)
done
(*>*)

subsection{*Derivation of the LOW rules*}

text{*We now derive the LOW rules according to the constructions given
in the paper.*}

lemma CAST[rule_format]:
  "G \<rhd> C : twiddle \<longrightarrow> G \<rhd> C : Sec (\<lambda> (s,t) . twiddle s t)"
(*<*)
apply clarsimp
apply (erule VDMConseq)  apply (simp add: twiddle_def Sec_def)
done
(*>*)

lemma SKIP: "G \<rhd> Skip : Sec (\<lambda> (s,t) . twiddle s t)"
(*<*)
apply (rule VDMConseq, rule VDMSkip)
apply (simp add: Sec_def twiddle_def)
done
(*>*)

lemma ASSIGN: 
  "(\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss) \<Longrightarrow>
   G \<rhd> (Assign x e) 
     : (Sec (\<lambda> (s, t) . twiddle s (update t x (evalE e t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: Sec_def)
  apply clarsimp 
  apply (erule_tac x=s in allE, erule_tac x=r in allE, clarsimp)
  apply (simp add: twiddle_def update_def)
done
(*>*)
(*
lemma ASSIGN: 
  "(e,low):VS_expr \<Longrightarrow>
    G \<rhd> (Assign x e) : (Sec (\<lambda> s ss . twiddle s (update ss x (evalE e ss))))"
  apply (drule ExprSound, clarsimp)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: Sec_def) 
  apply clarsimp  
  apply (erule_tac x=s in allE, erule_tac x=s2 in allE, clarsimp) 
  apply (simp add: twiddle_def update_def)
done
*)

lemma COMP: 
  "\<lbrakk> G \<rhd> C1 : (Sec phi1); G \<rhd> C2 : (Sec phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Comp C1 C2) : (Sec (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                           (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w)))))"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
    apply (simp add: Sec_def, clarsimp)
    apply rule
      apply clarsimp apply (erule_tac x=ra in allE, clarsimp)
        apply (rule_tac x=r in exI, clarsimp) 
     apply clarsimp
done
(*>*)

lemma IFF:
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss);
     G \<rhd> C1 : (Sec phi1); G \<rhd> C2 : (Sec phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Iff b C1 C2) : Sec (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                        ((\<not> evalB b t) \<longrightarrow> phi2(s,t)))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMIff) apply assumption+
  apply clarsimp apply (simp add: Sec_def)
  apply clarsimp 
  apply (case_tac "evalB b s")
  apply clarsimp 
  apply clarsimp 
done
(*>*)


text{*We introduce an explicit fixed point construction over the type
$TT$ of the invariants $\Phi$.*}

constdefs FIX::"(TT \<Rightarrow> TT) \<Rightarrow> TT"
"FIX Phi \<equiv> \<lambda> (s,t). \<forall> phi . (\<forall> ss tt . Phi phi (ss, tt) \<longrightarrow> phi (ss, tt))
                             \<longrightarrow> phi (s, t)" 

constdefs Monotone::"(TT \<Rightarrow> TT) \<Rightarrow> bool"
"Monotone Phi \<equiv> \<forall> phi psi . (\<forall> s t . phi(s,t) \<longrightarrow> psi(s,t)) \<longrightarrow> 
                        (\<forall> s t . Phi phi (s,t) \<longrightarrow> Phi psi (s,t))"

(*<*)
lemma Fix2: "\<lbrakk>Monotone Phi; Phi (FIX Phi) (s, t)\<rbrakk> \<Longrightarrow> FIX Phi (s,t)"
apply (simp add: FIX_def) apply clarsimp
apply (subgoal_tac "Phi phi (s,t)", simp)
apply (subgoal_tac "\<forall> r u . FIX Phi (r,u) \<longrightarrow> phi(r,u)")
prefer 2 apply (erule thin_rl) apply (simp add: FIX_def) apply clarsimp
  apply (erule_tac x=phi in allE, simp)
apply (unfold Monotone_def)
  apply (erule_tac x="FIX Phi" in allE, erule_tac x=phi in allE)
  apply (erule impE) apply assumption
  apply (simp add: FIX_def)
done

lemma Fix1: "\<lbrakk>Monotone Phi; FIX Phi (s,t)\<rbrakk> \<Longrightarrow> Phi (FIX Phi) (s,t)"
apply (simp add: FIX_def) 
apply (erule_tac x="Phi(FIX Phi)" in allE) 
apply (erule impE)
prefer 2 apply (simp add: FIX_def)
apply (subgoal_tac "\<forall> r u . Phi (FIX Phi) (r,u) \<longrightarrow> FIX Phi (r,u)")
  prefer 2 apply clarsimp apply (erule Fix2) apply assumption
apply (unfold Monotone_def)
  apply (erule_tac x="Phi(FIX Phi)" in allE, erule_tac x="FIX Phi" in allE, erule impE) apply assumption
apply simp
done
(*>*)

text{*For monotone assertion transformers, the construction indeed
yields a fixed point.*}

lemma Fix_lemma:"Monotone Phi \<Longrightarrow> Phi (FIX Phi) = FIX Phi"
(*<*)
apply rule
apply (case_tac x, clarsimp)
apply rule apply (erule Fix2, assumption) apply (erule Fix1, assumption) 
done
(*>*)

(*<*)
lemma "\<lbrakk>G \<rhd> C :Sec phi; G \<rhd> LOOP : Sec psi\<rbrakk> \<Longrightarrow> G \<rhd> (Iff b (Comp C LOOP) Skip) : Sec psi"
apply (rule VDMConseq)
apply (rule IFF) defer 1
apply (rule COMP)
apply assumption
apply assumption
apply (rule SKIP) apply clarsimp
apply (subgoal_tac "psi = (\<lambda>(s, t).
                   (evalB b t \<longrightarrow> (\<exists>r. phi (r, t) \<and> (\<forall>w. twiddle r w \<longrightarrow> psi (s, w)))) \<and> (\<not> evalB b t \<longrightarrow> twiddle s t))")
apply clarsimp
oops
(*>*)

text{*The following operator may be used to define a while rule using
fixed points.*}

constdefs PhiWhileOp::"BExpr \<Rightarrow> TT \<Rightarrow> TT \<Rightarrow> TT"
"PhiWhileOp b phi \<equiv> 
     \<lambda> psi . (\<lambda>(s, t).
             (evalB b t \<longrightarrow> (\<exists>r. phi (r, t) \<and> 
                                 (\<forall>w. twiddle r w \<longrightarrow> psi (s, w)))) \<and> 
             (\<not> evalB b t \<longrightarrow> twiddle s t))"

text{*Since this operator is monotone, \ldots*}

lemma PhiWhileOp_Monotone: "Monotone (PhiWhileOp b phi)"
(*<*)
apply (simp add: PhiWhileOp_def Monotone_def) apply clarsimp
  apply (rule_tac x=r in exI, simp)
done
(*>*)

text{*we may define its fixed point, *}

constdefs PhiWhile::"BExpr \<Rightarrow> TT \<Rightarrow> TT"
"PhiWhile b phi \<equiv> FIX (PhiWhileOp b phi)"

text{*which we can use to derive the following rule.*}

lemma WHILE:  
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss); 
     G \<rhd> C : (Sec phi)\<rbrakk> \<Longrightarrow>
   G \<rhd> (While b C) : (Sec (PhiWhile b phi))"
(*<*)
apply (rule VDMConseq)
apply (rule VDMWhile) 
prefer 4 apply (subgoal_tac "\<forall>s t. (Sec (PhiWhileOp b phi (PhiWhile b phi))) s t \<and> \<not> evalB b t \<longrightarrow> Sec (PhiWhile b phi) s t") apply assumption
  apply clarsimp apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi)= PhiWhile b phi", clarsimp)
                 apply (simp add: PhiWhile_def) apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
apply assumption
apply clarsimp apply (simp add: Sec_def) 
  apply (rule, clarsimp) apply (simp add: PhiWhileOp_def)
  apply clarsimp apply (simp add: PhiWhileOp_def)
apply clarsimp apply (simp add: Sec_def)
  apply rule
  prefer 2 apply clarsimp
    apply (subgoal_tac "\<exists>r. phi (r, s) \<and> (\<forall>w. twiddle r w \<longrightarrow> (PhiWhile b phi) (ra, w))")
    prefer 2 apply (simp add: PhiWhileOp_def) 
    apply clarsimp apply (rotate_tac -3, erule thin_rl)
    apply (rotate_tac -1, erule_tac x=ra in allE, erule mp)
    apply (rotate_tac 1, erule_tac x=r in allE, erule impE) apply fast
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
  apply clarsimp
    apply (simp (no_asm_simp) add: PhiWhileOp_def)
    apply (rule_tac x=r in exI, rule) apply simp
    apply clarsimp
    apply (rotate_tac 5, erule_tac x=w in allE, clarsimp)
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
done
(*>*)

text{*The operator that given $\phi$ returns the invariant
occurring in the conclusion of the rule is itself monotone - this is
the property required for the rule for procedure invocations.*}

lemma PhiWhileMonotone: "Monotone (\<lambda> phi . PhiWhile b phi)"
(*<*)
apply (simp add: Monotone_def) apply clarsimp
apply (simp add: PhiWhile_def)
apply (simp add: FIX_def) apply clarsimp
apply (erule_tac x=phia in allE, erule mp)
apply (clarsimp) apply (erule_tac x=ss in allE, erule_tac x=tt in allE, erule mp)
apply (simp add: PhiWhileOp_def) apply clarsimp
apply (rule_tac x=r in exI, simp)
done
(*>*)

text{*We now derive an alternative while rule that employs an
inductive formulation of a variant that replaces the fixed point
construction. This version is given in the paper.*}

text{*First, the inductive definition of the $\mathit{var}$ relation.*}
consts var::"(BExpr \<times> TT \<times> State \<times> State) set"
inductive var intros
varFalse: "\<lbrakk>\<not> evalB b t; twiddle s t\<rbrakk> \<Longrightarrow> (b,phi,s,t):var"
varTrue:  "\<lbrakk>evalB b t; phi(r,t); \<forall> w . twiddle r w \<longrightarrow> (b,phi,s,w): var\<rbrakk>
           \<Longrightarrow> (b,phi,s,t):var"

text{*It is easy to prove the equivalence of $\mathit{var}$ and the
fixed point:*}

lemma varFIX: "(b,phi,s,t):var \<Longrightarrow> PhiWhile b phi (s,t)"
(*<*)
apply (erule var.induct)
apply (simp add: PhiWhile_def)
  apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) (s,t)")
  apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) = FIX (PhiWhileOp b phi)", clarsimp)
  apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
  apply (simp add: PhiWhileOp_def)
apply (simp (no_asm_simp) add: PhiWhile_def)
apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) (s,t)")
  apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) = FIX (PhiWhileOp b phi)", clarsimp)
  apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
  apply (simp add: PhiWhileOp_def)
  apply (rule_tac x=r in exI, simp)
  apply clarsimp
  apply (erule_tac x=w in allE, clarsimp)
  apply (simp add: PhiWhile_def)
  apply (simp add: PhiWhileOp_def)
done

lemma FIXvar: "PhiWhile b phi (s,t) \<Longrightarrow> (b,phi,s,t):var"
apply (simp add: PhiWhile_def)
apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) (s, t)")
prefer 2 
  apply (subgoal_tac "PhiWhileOp b phi (FIX (PhiWhileOp b phi)) = FIX (PhiWhileOp b phi)", clarsimp)
  apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
apply (erule thin_rl, simp add: PhiWhileOp_def) apply clarsimp
  apply (case_tac "evalB b t")
  prefer 2 apply clarsimp apply (rule varFalse) apply assumption+
  apply clarsimp apply (rule varTrue) apply assumption apply assumption 
    apply clarsimp apply (erule_tac x=w in allE, clarsimp)
    apply (unfold FIX_def) apply clarify
    apply (erule_tac x="\<lambda> (x,y) . (b,phi,x,y):var" in allE, erule impE) prefer 2 apply simp
    apply clarsimp
    apply (case_tac "evalB b tt")
    prefer 2 apply clarsimp apply (rule varFalse) apply assumption+
    apply clarsimp apply (rule varTrue) apply assumption+
done

lemma varFIXvar: "(PhiWhile b phi (s,t)) = ((b,phi,s,t):var)"
apply rule
apply (erule FIXvar)
apply (erule varFIX)
done

lemma FIXvarFIX: "(PhiWhile b phi) = (\<lambda> (s,t) . (b,phi,s,t):var)"
apply rule
apply rule
apply (case_tac x, clarsimp)
apply (erule FIXvar)
apply (case_tac x, clarsimp)
apply (erule varFIX)
done
(*>*)

lemma FIXvarFIX1: "(PhiWhile b) = (\<lambda> phi . (\<lambda> (s,t) . (b,phi,s,t):var))"
(*<*)
by (rule, rule FIXvarFIX)
(*>*)

text{*Thus, we obtain the while rule given in the paper from the rule
we gave above.*}

lemma WHILE_IND:
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss); 
     G \<rhd> C : (Sec phi)\<rbrakk> \<Longrightarrow>
   G \<rhd> (While b C) : (Sec (\<lambda> (s,t) . (b,phi,s,t):var))"
(*<*)
apply (rule VDMConseq)
apply (rule WHILE) apply assumption+
apply clarsimp
apply (simp add: FIXvarFIX)
done
(*>*)

text{*Not suprisingly, the construction $\mathit{var}$ can be shown to
be monotone in $\phi$.*}

(*<*)
lemma varMonotoneAux[rule_format]:
  "(b, phi, s, t) \<in> var \<Longrightarrow> 
   (\<forall>s t. phi (s, t) \<longrightarrow> psi (s, t)) \<longrightarrow> (b, psi, s, t) \<in> var"
apply (erule var.induct)
apply clarsimp apply (erule varFalse, simp)
apply clarsimp apply (erule varTrue) apply fast apply simp
done
(*>*)

lemma var_Monotone: "Monotone (\<lambda> phi . (\<lambda> (s,t) .(b,phi,s,t):var))"
(*<*)
apply (simp add: Monotone_def)
apply clarsimp
apply (rule varMonotoneAux) apply assumption apply simp
done
(*>*)
(*<*)
lemma varMonotone_byFIX: "Monotone (\<lambda> phi . (\<lambda> (s,t) .(b,phi,s,t):var))"
apply (subgoal_tac "Monotone (\<lambda> phi . PhiWhile b phi)")
apply (simp add: FIXvarFIX1)
apply (rule PhiWhileMonotone)
done  
(*>*)

text{*The call rule is now formulated for an arbitrary fixed point of
a monotone transformer.*}

lemma CALL: 
  "\<lbrakk> ({Sec(FIX Phi)} \<union> G) \<rhd> body : Sec(Phi(FIX Phi));
     Monotone Phi\<rbrakk> \<Longrightarrow>
   G \<rhd> Call : Sec(FIX Phi)"
(*<*)
apply (rule VDMCall)
apply (subgoal_tac "Phi (FIX Phi) = FIX Phi", clarsimp)
apply (erule Fix_lemma)
done
(*>*)

subsection{*Derivation of the HIGH rules*}

text{*The HIGH rules are easy.*}

lemma HIGH_SKIP: "G \<rhd> Skip : twiddle"
(*<*)
apply (rule VDMConseq, rule VDMSkip) 
apply (simp add: twiddle_def)
done
(*>*)

lemma HIGH_ASSIGN:
  "CONTEXT x = high \<Longrightarrow> G \<rhd> (Assign x e) : twiddle"
(*<*)
apply (rule VDMConseq, rule VDMAssign) 
apply (simp add: update_def twiddle_def)
done
(*>*)

lemma HIGH_COMP: 
  "\<lbrakk> G \<rhd> C1 : twiddle; G \<rhd> C2 : twiddle\<rbrakk> 
  \<Longrightarrow> G \<rhd> (Comp C1 C2): twiddle"
(*<*)
apply (rule VDMConseq, rule VDMComp) apply assumption+
apply (simp add: twiddle_def)
apply clarsimp
done
(*>*)

lemma HIGH_IF:
  "\<lbrakk> G \<rhd> C1 : twiddle; G \<rhd> C2 : twiddle \<rbrakk> 
  \<Longrightarrow> G \<rhd> (Iff b C1 C2) : twiddle"
(*<*)
apply (rule VDMConseq, rule VDMIff) apply assumption+
apply (simp add: twiddle_def)
done
(*>*)

lemma HIGH_WHILE:
  "\<lbrakk> G \<rhd> C  : twiddle\<rbrakk> \<Longrightarrow> G \<rhd> (While b C)  : twiddle"
(*<*)
apply (rule VDMConseq)
  apply (rule VDMWhile) 
  apply (subgoal_tac "G \<rhd> C : (\<lambda>s t. evalB b s \<longrightarrow> twiddle s t)", erule thin_rl, assumption)
  apply (erule VDMConseq) apply simp
  prefer 3 apply (subgoal_tac "\<forall>s t. twiddle s t \<and> \<not> evalB b t \<longrightarrow> twiddle s t", assumption) apply simp
  apply (simp add: twiddle_def)
  apply (simp add: twiddle_def)
done
(*>*)

lemma HIGH_CALL:
  "({twiddle} \<union> G) \<rhd> body : twiddle \<Longrightarrow> G \<rhd> Call : twiddle"
(*<*)
by (erule VDMCall)
(*>*)

subsection{*The type system of Volpano, Smith and Irvine*}

text{*First, type systems for expressions and boolean expressions.*}

consts VS_expr:: "(Expr \<times> TP) set"
inductive VS_expr intros
VS_exprVar: "CONTEXT x = t \<Longrightarrow> (varE x, t) : VS_expr"
VS_exprVal: "(valE v, low) : VS_expr"
VS_exprOp: "\<lbrakk>(e1,t) : VS_expr; (e2,t):VS_expr\<rbrakk>
             \<Longrightarrow> (opE f e1 e2,t) : VS_expr"
VS_exprHigh: "(e, high) : VS_expr"

consts VS_Bexpr:: "(BExpr \<times> TP) set"
inductive VS_Bexpr intros
VS_BexprOp: "\<lbrakk>(e1,t) : VS_expr; (e2,t):VS_expr\<rbrakk>
             \<Longrightarrow> (compB f e1 e2,t) : VS_Bexpr"
VS_BexprHigh: "(e,high) : VS_Bexpr"

text{*Next, the core of the type system, the rules for commands.*}

consts VS_com:: "(TP \<times> IMP) set"
inductive VS_com intros

VS_comSkip: "(pc,Skip) : VS_com"

VS_comAssHigh: "CONTEXT x = high \<Longrightarrow> (pc,Assign x e) : VS_com"

VS_comAssLow: "\<lbrakk>CONTEXT x = low; pc = low; (e,low):VS_expr\<rbrakk>
               \<Longrightarrow> (pc,Assign x e) : VS_com"

VS_comComp: "\<lbrakk> (pc,c1):VS_com; (pc,c2):VS_com\<rbrakk> 
             \<Longrightarrow> (pc,Comp c1 c2) : VS_com"

VS_comIf: "\<lbrakk> (b,pc):VS_Bexpr; (pc,c1):VS_com; (pc,c2):VS_com\<rbrakk>
           \<Longrightarrow> (pc,Iff b c1 c2):VS_com"

VS_comWhile: "\<lbrakk>(b,pc):VS_Bexpr; (pc,c):VS_com\<rbrakk>
              \<Longrightarrow> (pc,While b c):VS_com"

VS_comSub: "(high,c) : VS_com \<Longrightarrow> (low,c):VS_com"

text{*We define the interpretation of expression typings\ldots *}

consts SemExpr::"Expr \<Rightarrow> TP \<Rightarrow> bool" 
primrec
"SemExpr e low = (\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss)"
"SemExpr e high = True"

text{*\ldots and show the soundness of the typing rules.*}

lemma ExprSound: "(e,tp):VS_expr \<Longrightarrow> SemExpr e tp"
(*<*)
apply (erule VS_expr.induct)
apply (case_tac t,simp_all)
  apply (simp add: twiddle_def)
apply (case_tac t, simp_all)
done
(*>*)

text{*Likewise for the boolean expressions.*}

consts SemBExpr::"BExpr \<Rightarrow> TP \<Rightarrow> bool"
primrec
"SemBExpr b low = (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss)"
"SemBExpr b high = True"

lemma BExprSound: "(e,tp):VS_Bexpr \<Longrightarrow> SemBExpr e tp"
(*<*)
apply (erule VS_Bexpr.induct, simp_all)
apply (drule ExprSound) 
apply (drule ExprSound) 
apply (case_tac t, simp_all) 
done
(*>*)

text{*The proof of the main theorem proceeds by induction on
$(t,C):VS\_com$.*}

theorem VS_com_VDM[rule_format]:
"(t,C):VS_com \<Longrightarrow> (t=high \<longrightarrow> G \<rhd> C : twiddle) \<and>
                  (t=low \<longrightarrow> (\<exists> A . G \<rhd> C : Sec A))"
(*<*)
apply (erule VS_com.induct)
apply rule
  apply clarsimp
  apply (rule HIGH_SKIP)
  apply clarsimp
  apply (rule, rule SKIP)
apply rule
  apply clarsimp
  apply (erule HIGH_ASSIGN)
  apply clarsimp
  apply (subgoal_tac "\<exists> t. (e,t):VS_expr", clarsimp) prefer 2 apply (rule, rule VS_exprHigh)
  apply (drule ExprSound)
  apply (case_tac t)
    apply clarsimp apply (rule, erule ASSIGN) 
    apply clarsimp apply (rule_tac x="\<lambda> (s,t).  twiddle s t" in exI, rule VDMConseq) 
         apply (erule HIGH_ASSIGN) apply (simp add: Sec_def twiddle_def)
apply rule
  apply clarsimp
  apply (drule ExprSound)
    apply clarsimp apply (rule, erule ASSIGN) 
apply rule
  apply clarsimp
  apply (rule HIGH_COMP) apply (assumption, assumption)
  apply clarsimp
  apply (rule, rule COMP) apply (assumption, assumption)
apply rule
  apply clarsimp
  apply (rule HIGH_IF) apply (assumption, assumption)
  apply clarsimp
  apply (drule BExprSound)
  apply clarsimp
  apply (rule, erule IFF)  apply (assumption, assumption)
apply rule
  apply clarsimp
  apply (erule HIGH_WHILE) 
  apply clarsimp
  apply (drule BExprSound)
  apply clarsimp
  apply (rule, erule WHILE) apply assumption
apply clarsimp
  apply (rule, erule CAST)
done
(*>*)

text{*The semantic of typing judgements for commands is now the
expected one: HIGH commands require initial and final state be low
equivalent (i.e.~the low variables in the final state can't depend on
the high variables of the initial state), while LOW commands must
respect the above mentioned security property.*}

consts SemCom::"TP \<Rightarrow> IMP \<Rightarrow> bool"
primrec
"SemCom low c = secure c"
"SemCom high c = (\<forall> s t . (s,c \<Down> t) \<longrightarrow> twiddle s t)"

text{*By induction on the typing rules for commands, we obtain the
soundness result.*}

theorem VS_SOUND: "(t,C):VS_com \<Longrightarrow> SemCom t C"
(*<*)
apply (subgoal_tac "(t=high \<longrightarrow> {} \<rhd> C : twiddle) \<and> (t=low \<longrightarrow> (\<exists> A . {} \<rhd> C : Sec A))")
prefer 2 apply (erule VS_com_VDM)
apply (case_tac t)
apply clarsimp apply (drule VDM_Sound) apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def Sec_def secure_def) 
apply clarsimp apply (drule VDM_Sound) apply (simp add: v_valid_def vdmctxt_valid_def vdm_valid_def Sec_def secure_def) 
done
(*>*)

(*<*)
consts Deriv::"(VDMAssn set \<times> IMP \<times> VDMAssn) set"
inductive Deriv intros
D_CAST: "(G,C,twiddle):Deriv \<Longrightarrow> (G, C, Sec (\<lambda> (s,t) . twiddle s t)) : Deriv"
D_SKIP: "(G, Skip, Sec (\<lambda> (s,t) . twiddle s t)) : Deriv"
D_ASSIGN: "(\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss) \<Longrightarrow>
   (G, Assign x e, Sec (\<lambda> (s, t) . twiddle s (update t x (evalE e t)))):Deriv"
D_COMP: 
  "\<lbrakk> (G, C1, Sec phi1):Deriv; (G, C2, Sec phi2):Deriv\<rbrakk> \<Longrightarrow>
   (G, Comp C1 C2, Sec (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                           (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w))))):Deriv"
C_IFF:
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss);
     (G, C1, Sec phi1):Deriv; (G,C2, Sec phi2):Deriv\<rbrakk> \<Longrightarrow>
     (G, Iff b C1 C2, Sec (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                        ((\<not> evalB b t) \<longrightarrow> phi2(s,t)))):Deriv"
D_WHILE:  
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss); 
     (G, C, Sec phi):Deriv\<rbrakk> \<Longrightarrow>
   (G, While b C, Sec (\<lambda> (s,t) . (b,phi,s,t):var)):Deriv"
D_CALL:
  "\<lbrakk> ({Sec(FIX Phi)} \<union> G, body, Sec(Phi(FIX Phi))):Deriv;
     Monotone Phi\<rbrakk> \<Longrightarrow>
    (G, Call, Sec(FIX Phi)):Deriv"

constdefs DProp :: "VDMAssn \<Rightarrow> bool"
"DProp A \<equiv> (\<exists> Phi . A =  Sec (Phi (FIX Phi)) \<and> Monotone Phi)"

lemma DerivProp_Aux: "(X,C,A):Deriv \<Longrightarrow> DProp A"
apply (erule Deriv.induct)
apply (simp_all add: DProp_def)
  apply clarsimp
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
  apply clarsimp ?
  apply (rule, rule) apply simp apply (simp add: Monotone_def)
done

lemma DerivProp: 
 "(X,C,A):Deriv
  \<Longrightarrow> \<exists> Q R S Phi . A =  Sec (Phi (FIX Phi)) \<and> Monotone Phi"
by (drule DerivProp_Aux, simp add: DProp_def)
(*>*)


subsection{*Derivation of VSI in separated form*} 

text{*In this subsection, we repeat the embedding of VSI's rules but
use the separation into left and right rules from the new paper.*}

constdefs SecL :: "TT \<Rightarrow> VDMAssn"
"SecL phi s t \<equiv> \<forall> r . twiddle s r \<longrightarrow> phi(t, r)"

constdefs SecR :: "TT \<Rightarrow> VDMAssn"
"SecR phi s t \<equiv> \<forall> r . phi(r ,s) \<longrightarrow> twiddle r t"

text{*First, just a variation of some rules where we unfold the
separation*}

constdefs INTER:: "VDMAssn \<Rightarrow> VDMAssn \<Rightarrow> VDMAssn"
"INTER A B \<equiv> \<lambda> s t . A s t \<and> B s t"

lemma Sec_SecL_SecR:
"Sec phi = INTER (SecL phi) (SecR phi)"
(*<*)
apply (rule, rule) apply (simp add: INTER_def Sec_def SecL_def SecR_def) 
done
(*>*)

lemma LR_CAST[rule_format]:
  "G \<rhd> C : twiddle \<Longrightarrow> 
  G \<rhd> C : INTER (SecL (\<lambda> (s,t) . twiddle s t)) 
                 (SecR (\<lambda> (s,t) . twiddle s t))"
(*<*)
apply (erule VDMConseq)
apply (simp add: twiddle_def INTER_def SecL_def SecR_def)
done
(*>*)

lemma LR_SKIP: "G \<rhd> Skip : INTER (SecL (\<lambda> (s,t) . twiddle s t))
                                 (SecR (\<lambda> (s,t) . twiddle s t))"
(*<*)
apply (rule VDMConseq, rule VDMSkip)
apply (simp add: INTER_def SecL_def SecR_def twiddle_def)
done
(*>*)

lemma LR_ASSIGN: 
  "(\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss) \<Longrightarrow>
   G \<rhd> (Assign x e) 
     : INTER (SecL (\<lambda> (s, t) . twiddle s (update t x (evalE e t))))
             (SecR (\<lambda> (s, t) . twiddle s (update t x (evalE e t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: INTER_def SecL_def SecR_def)
  apply clarsimp 
  apply (erule_tac x=s in allE, erule_tac x=r in allE, clarsimp)
  apply (simp add: twiddle_def update_def)
done
(*>*)
(*
lemma ASSIGN: 
  "(e,low):VS_expr \<Longrightarrow>
    G \<rhd> (Assign x e) : (Sec (\<lambda> s ss . twiddle s (update ss x (evalE e ss))))"
  apply (drule ExprSound, clarsimp)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: Sec_def) 
  apply clarsimp  
  apply (erule_tac x=s in allE, erule_tac x=s2 in allE, clarsimp) 
  apply (simp add: twiddle_def update_def)
done
*)

lemma LR_COMP: 
  "\<lbrakk> G \<rhd> C1 : INTER (SecL phi1) (SecR phi1);
     G \<rhd> C2 : INTER (SecL phi2) (SecR phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Comp C1 C2) : INTER (SecL (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                                    (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w)))))
                               (SecR (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                                    (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w)))))"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
    apply (simp add: INTER_def SecL_def SecR_def, clarsimp)
    apply rule
      apply clarsimp apply (erule_tac x=ra in allE, clarsimp)
        apply (rule_tac x=r in exI, clarsimp) 
     apply clarsimp
done
(*>*)

lemma LR_IFF:
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss);
     G \<rhd> C1 : INTER (SecL phi1) (SecR phi1);
     G \<rhd> C2 : INTER (SecL phi2) (SecR phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Iff b C1 C2) : INTER (SecL (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                        ((\<not> evalB b t) \<longrightarrow> phi2(s,t))))
                                (SecR (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                        ((\<not> evalB b t) \<longrightarrow> phi2(s,t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMIff) apply assumption+
  apply clarsimp apply (simp add: INTER_def SecL_def SecR_def)
  apply clarsimp 
  apply (case_tac "evalB b s")
  apply clarsimp 
  apply clarsimp 
done
(*>*)

lemma LR_WHILE:  
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss); 
     G \<rhd> C : INTER (SecL phi) (SecR phi)\<rbrakk> \<Longrightarrow>
   G \<rhd> (While b C) : INTER (SecL (PhiWhile b phi)) (SecR (PhiWhile b phi))"
(*<*)
apply (rule VDMConseq)
apply (rule VDMWhile) 
prefer 4 apply (subgoal_tac "\<forall>s t. INTER (SecL (PhiWhileOp b phi (PhiWhile b phi))) (SecR (PhiWhileOp b phi (PhiWhile b phi))) s t \<and> \<not> evalB b t \<longrightarrow> INTER (SecL (PhiWhile b phi)) (SecR (PhiWhile b phi)) s t") apply assumption
  apply clarsimp apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi)= PhiWhile b phi", clarsimp)
                 apply (simp add: PhiWhile_def) apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
apply assumption
apply clarsimp apply (simp add: INTER_def SecL_def SecR_def) 
  apply (rule, clarsimp) apply (simp add: PhiWhileOp_def)
  apply clarsimp apply (simp add: PhiWhileOp_def)
apply clarsimp apply (simp add: INTER_def SecL_def SecR_def)
  apply rule
  prefer 2 apply clarsimp
    apply (subgoal_tac "\<exists>r. phi (r, s) \<and> (\<forall>w. twiddle r w \<longrightarrow> (PhiWhile b phi) (ra, w))")
    prefer 2 apply (simp add: PhiWhileOp_def) 
    apply clarsimp apply (rotate_tac -3, erule thin_rl)
    apply (rotate_tac -1, erule_tac x=ra in allE, erule mp)
    apply (rotate_tac 1, erule_tac x=r in allE, erule impE) apply fast
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
  apply clarsimp
    apply (simp (no_asm_simp) add: PhiWhileOp_def)
    apply (rule_tac x=r in exI, rule) apply simp
    apply clarsimp
    apply (rotate_tac 5, erule_tac x=w in allE, clarsimp)
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
done
(*>*)

lemma LR_WHILE_IND:
  "\<lbrakk> (\<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss); 
     G \<rhd> C : INTER (SecL phi) (SecR phi)\<rbrakk> \<Longrightarrow>
   G \<rhd> (While b C) : INTER (SecL (\<lambda> (s,t) . (b,phi,s,t):var)) (SecR (\<lambda> (s,t) . (b,phi,s,t):var))"
(*<*)
apply (rule VDMConseq)
apply (rule LR_WHILE) apply assumption+
apply clarsimp
apply (simp add: FIXvarFIX)
done
(*>*)

lemma LR_CALL: 
  "\<lbrakk> ({INTER (SecL (FIX Phi)) (SecR (FIX Phi))} \<union> G) \<rhd> body : INTER (SecL (Phi(FIX Phi))) (SecR (Phi(FIX Phi)));
     Monotone Phi\<rbrakk> \<Longrightarrow>
   G \<rhd> Call : INTER (SecL (FIX Phi)) (SecR (FIX Phi))"
(*<*)
apply (rule VDMCall)
apply (subgoal_tac "Phi (FIX Phi) = FIX Phi", clarsimp)
apply (erule Fix_lemma)
done
(*>*)

text{*Next, the separate rules for SecL and SecR.*}

lemma LRsepL_CAST[rule_format]:
  "G \<rhd> C : twiddle \<Longrightarrow> 
  G \<rhd> C : (SecL (\<lambda> (s,t) . twiddle s t)) "
(*<*)
apply (erule VDMConseq)
apply (simp add: twiddle_def INTER_def SecL_def SecR_def)
done
(*>*)

lemma LRsepR_CAST[rule_format]:
  "G \<rhd> C : twiddle \<Longrightarrow> 
  G \<rhd> C : (SecR (\<lambda> (s,t) . twiddle s t)) "
(*<*)
apply (erule VDMConseq)
apply (simp add: twiddle_def INTER_def SecL_def SecR_def)
done
(*>*)

lemma LRsepL_SKIP: "G \<rhd> Skip : (SecL (\<lambda> (s,t) . twiddle s t))"
(*<*)
apply (rule VDMConseq, rule VDMSkip)
apply (simp add: INTER_def SecL_def SecR_def twiddle_def)
done
(*>*)

lemma LRsepR_SKIP: "G \<rhd> Skip : (SecR (\<lambda> (s,t) . twiddle s t))"
(*<*)
apply (rule VDMConseq, rule VDMSkip)
apply (simp add: INTER_def SecL_def SecR_def twiddle_def)
done
(*>*)

lemma LRsepL_ASSIGN: 
  "(\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss) \<Longrightarrow>
   G \<rhd> (Assign x e) 
     : (SecL (\<lambda> (s, t) . twiddle s (update t x (evalE e t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: INTER_def SecL_def SecR_def)
  apply clarsimp 
  apply (erule_tac x=s in allE, erule_tac x=r in allE, clarsimp)
  apply (simp add: twiddle_def update_def)
done
(*>*)
lemma LRsepR_ASSIGN: 
  "(\<forall> s ss. twiddle s ss \<longrightarrow> evalE e s = evalE e ss) \<Longrightarrow>
   G \<rhd> (Assign x e)
     : (SecR (\<lambda> (s, t) . twiddle s (update t x (evalE e t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: INTER_def SecL_def SecR_def)
done
(*>*)
(*
lemma ASSIGN: 
  "(e,low):VS_expr \<Longrightarrow>
    G \<rhd> (Assign x e) : (Sec (\<lambda> s ss . twiddle s (update ss x (evalE e ss))))"
  apply (drule ExprSound, clarsimp)
 apply (rule VDMConseq) apply(rule VDMAssign)
  apply clarsimp apply (simp add: Sec_def) 
  apply clarsimp  
  apply (erule_tac x=s in allE, erule_tac x=s2 in allE, clarsimp) 
  apply (simp add: twiddle_def update_def)
done
*)

text{*In the rule for composition, we observe for the first time the
\emph{perfect separation}: for deriving the statement for SecL we only
need assumptions involving SecL, and similarly for SecR.*}

lemma LRsepL_COMP: 
  "\<lbrakk> G \<rhd> C1 : (SecL phi1); G \<rhd> C2 : (SecL phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Comp C1 C2) : (SecL (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                                    (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w)))))"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
    apply (simp add: SecL_def, clarsimp)
    apply (erule_tac x=ra in allE, clarsimp)
        apply (rule_tac x=r in exI, clarsimp) 
done
(*>*)

lemma LRsepR_COMP: 
  "\<lbrakk> G \<rhd> C1 : (SecR phi1);
     G \<rhd> C2 : (SecR phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Comp C1 C2) : (SecR (\<lambda> (s,t) . \<exists> r . phi1(r, t) \<and> 
                                    (\<forall> w . (twiddle r w \<longrightarrow> phi2(s, w)))))"
(*<*)
apply (rule VDMConseq) apply (erule VDMComp) apply assumption
    apply (simp add: SecR_def, clarsimp)
done
(*>*)

text{*Similarly for conditionals*}

lemma LRsepL_IFF:
  "\<lbrakk> G \<rhd> C1 : (SecL phi1); G \<rhd> C2 : (SecL phi2);
     \<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss\<rbrakk>
  \<Longrightarrow> G \<rhd> (Iff b C1 C2) : (SecL (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                             ((\<not> evalB b t) \<longrightarrow> phi2(s,t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMIff) apply assumption+
  apply clarsimp apply (simp add: SecL_def)
done
(*>*)

text{*In fact, the rule for SecR can be proven without recourse to the
side condition concerning the branch condition.*}

lemma LRsepR_IFF:
  "\<lbrakk> G \<rhd> C1 : (SecR phi1); G \<rhd> C2 : (SecR phi2)\<rbrakk> \<Longrightarrow>
     G \<rhd> (Iff b C1 C2) : (SecR (\<lambda> (s, t) . (evalB b t \<longrightarrow> phi1(s,t)) \<and> 
                                        ((\<not> evalB b t) \<longrightarrow> phi2(s,t))))"
(*<*)
 apply (rule VDMConseq) apply(rule VDMIff) apply assumption+
  apply clarsimp apply (simp add: SecR_def)
  apply clarsimp 
  apply (case_tac "evalB b s")
  apply clarsimp 
  apply clarsimp 
done
(*>*)

text{*Perfect separation also holds for loops.*}

lemma LRsepL_WHILE:  
  "\<lbrakk> G \<rhd> C : (SecL phi);
    \<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss\<rbrakk>
  \<Longrightarrow> G \<rhd> (While b C) : (SecL (PhiWhile b phi))"
(*<*)
apply (rule VDMConseq)
apply (rule VDMWhile) 
prefer 4 apply (subgoal_tac "\<forall>s t. (SecL (PhiWhileOp b phi (PhiWhile b phi))) s t \<and> \<not> evalB b t \<longrightarrow> (SecL (PhiWhile b phi)) s t") apply assumption
  apply clarsimp apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi)= PhiWhile b phi", clarsimp)
                 apply (simp add: PhiWhile_def) apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
apply assumption
apply clarsimp apply (simp add: SecL_def) 
  apply (rule, clarsimp) apply (simp add: PhiWhileOp_def)
apply clarsimp apply (simp add: SecL_def)
  apply clarsimp
    apply (simp (no_asm_simp) add: PhiWhileOp_def)
    apply (rule_tac x=r in exI, rule) apply simp
    apply clarsimp
    apply (rotate_tac 4, erule_tac x=w in allE, clarsimp)
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
done
(*>*)

text{*Again, the side condition on the boolean condition is not
reqquired for the rule for SecR.*}

lemma LRsepR_WHILE:  
  "G \<rhd> C : (SecR phi) \<Longrightarrow>
   G \<rhd> (While b C) : (SecR (PhiWhile b phi))"
(*<*)
apply (rule VDMConseq)
apply (rule VDMWhile) 
prefer 4 apply (subgoal_tac "\<forall>s t. (SecR (PhiWhileOp b phi (PhiWhile b phi))) s t \<and> \<not> evalB b t \<longrightarrow> SecR (PhiWhile b phi) s t") apply assumption
  apply clarsimp apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi)= PhiWhile b phi", clarsimp)
                 apply (simp add: PhiWhile_def) apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
apply assumption
apply clarsimp apply (simp add: SecR_def) 
  apply clarsimp apply (simp add: PhiWhileOp_def)
apply clarsimp apply (simp add: SecR_def)
  apply clarsimp
    apply (subgoal_tac "\<exists>r. phi (r, s) \<and> (\<forall>w. twiddle r w \<longrightarrow> (PhiWhile b phi) (ra, w))")
    prefer 2 apply (simp add: PhiWhileOp_def) 
    apply clarsimp apply (rotate_tac -3, erule thin_rl)
    apply (rotate_tac -1, erule_tac x=ra in allE, erule mp)
    apply (rotate_tac 1, erule_tac x=r in allE, erule impE) apply fast
    apply (subgoal_tac "PhiWhileOp b phi (PhiWhile b phi) = PhiWhile b phi", clarsimp)
    apply (simp add: PhiWhile_def)
    apply (rule Fix_lemma) apply (rule PhiWhileOp_Monotone)
done
(*>*)

text{*Here are the separated loop rules for the inductive
presentation.*}

lemma LRsepL_WHILE_IND:
  "\<lbrakk> G \<rhd> C : (SecL phi);
     \<forall> s ss. twiddle s ss \<longrightarrow> evalB b s = evalB b ss \<rbrakk>
  \<Longrightarrow> G \<rhd> (While b C) : (SecL (\<lambda> (s,t) . (b,phi,s,t):var))"
(*<*)
apply (rule VDMConseq)
apply (rule LRsepL_WHILE) apply assumption+
apply clarsimp
apply (simp add: FIXvarFIX)
done
(*>*)

lemma LRsepR_WHILE_IND:
  "G \<rhd> C : (SecR phi) \<Longrightarrow>
   G \<rhd> (While b C) : (SecR (\<lambda> (s,t) . (b,phi,s,t):var))"
(*<*)
apply (rule VDMConseq)
apply (erule LRsepR_WHILE)
apply clarsimp
apply (simp add: FIXvarFIX)
done
(*>*)

text{*Finally, the rules for the procedure.*}

lemma LRsepL_CALL: 
  "\<lbrakk> ({(SecL (FIX Phi))} \<union> G) \<rhd> body : (SecL (Phi(FIX Phi)));
     Monotone Phi\<rbrakk> \<Longrightarrow>
   G \<rhd> Call : (SecL (FIX Phi))"
(*<*)
apply (rule VDMCall)
apply (subgoal_tac "Phi (FIX Phi) = FIX Phi", clarsimp)
apply (erule Fix_lemma)
done
(*>*)

lemma LRsepR_CALL: 
  "\<lbrakk> ({(SecR (FIX Phi))} \<union> G) \<rhd> body : (SecR (Phi(FIX Phi)));
     Monotone Phi\<rbrakk> \<Longrightarrow>
   G \<rhd> Call : (SecR (FIX Phi))"
(*<*)
apply (rule VDMCall)
apply (subgoal_tac "Phi (FIX Phi) = FIX Phi", clarsimp)
apply (erule Fix_lemma)
done
(*>*)

text{*We already proved the VDM rule for intersecting specifications,
but the proofs above also yield the following semantic separation
result for security.*}

lemma "\<lbrakk>{} \<rhd> C : (SecL phi1); {} \<rhd> C : (SecR phi1)\<rbrakk> \<Longrightarrow> secure C"
(*<*)
apply (drule VDM_Sound)+
apply (simp add: v_valid_def secure_def)
apply (erule impE, simp add: vdmctxt_valid_def)+
apply (simp add: vdm_valid_def SecL_def SecR_def)
done
(*>*)

(*<*)
end
(*>*)
